from typing                                                                                       import Union, Dict
from fastapi                                                                                      import HTTPException, Response, Body
from osbot_fast_api.api.decorators.route_path                                                     import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                   import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                        import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                           import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                                 import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                 import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                                import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                   import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                    import type_safe
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Retrieve__Request     import Schema__Cache__Data__Retrieve__Request
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Retrieve__Response    import Schema__Cache__Data__Retrieve__Response
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                    import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                                import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                         import Cache__Service
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__Retrieve                    import Cache__Service__Data__Retrieve

TAG__ROUTES_RETRIEVE__DATA       = Safe_Str__Fast_API__Route__Tag('data')
PREFIX__ROUTES_RETRIEVE__DATA    = Safe_Str__Fast_API__Route__Prefix('/{namespace}/cache/{cache_id}')
BASE_PATH__ROUTES_RETRIEVE__DATA = f'{PREFIX__ROUTES_RETRIEVE__DATA}/{TAG__ROUTES_RETRIEVE__DATA}/'

ROUTES_PATHS__RETRIEVE__DATA = [ BASE_PATH__ROUTES_RETRIEVE__DATA + 'json/{data_file_id}'                  ,
                                 BASE_PATH__ROUTES_RETRIEVE__DATA + 'json/{data_key:path}/{data_file_id}'  ,
                                 BASE_PATH__ROUTES_RETRIEVE__DATA + 'string/{data_file_id}'                ,
                                 BASE_PATH__ROUTES_RETRIEVE__DATA + 'string/{data_key:path}/{data_file_id}',
                                 BASE_PATH__ROUTES_RETRIEVE__DATA + 'binary/{data_file_id}'                ,
                                 BASE_PATH__ROUTES_RETRIEVE__DATA + 'binary/{data_key:path}/{data_file_id}']


class Routes__Data__Retrieve(Fast_API__Routes):                                                 # FastAPI routes for retrieving data files from cache entries
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_RETRIEVE__DATA
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_RETRIEVE__DATA
    cache_service : Cache__Service                                                              # Dependency injection for cache service

    @cache_on_self
    def retrieve_service(self) -> Cache__Service__Data__Retrieve:                               # Service layer for data retrieval logic
        return Cache__Service__Data__Retrieve(cache_service=self.cache_service)


    @route_path("/data/json/{data_file_id}")
    def data__json__with__id(self, cache_id     : Cache_Id             = None                          ,
                                   namespace    : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                   data_file_id : Safe_Str__Id         = None
                             ) -> dict:
        return self.data__json__with__id_and_key(cache_id     = cache_id    ,
                                                 namespace    = namespace   ,
                                                 data_key     = ''          ,
                                                 data_file_id = data_file_id)

    @route_path("/data/json/{data_key:path}/{data_file_id}")
    def data__json__with__id_and_key(self, cache_id     : Cache_Id             = None                          ,
                                           namespace    : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                           data_key     : Safe_Str__File__Path = None                          ,
                                           data_file_id : Safe_Str__Id         = None
                                     ) -> dict:                                                                 # Retrieve data as JSON format
        request = Schema__Cache__Data__Retrieve__Request(cache_id     = cache_id              ,
                                                         data_type    = Enum__Cache__Data_Type.JSON,
                                                         data_key     = data_key              ,
                                                         data_file_id = data_file_id          ,
                                                         namespace    = namespace             )

        result = self.retrieve_service().retrieve_data(request)

        return self.handle_json_result(result, cache_id=cache_id, data_file_id=data_file_id)

    @route_path("/data/string/{data_file_id}")
    def data__string__with__id(self, cache_id     : Cache_Id             = None                          ,
                                     namespace    : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                     data_file_id : Safe_Str__Id         = None
                               ) -> Response:
        return self.data__string__with__id_and_key(cache_id     = cache_id    ,
                                                   namespace    = namespace   ,
                                                   data_key     = ''          ,
                                                   data_file_id = data_file_id)

    @route_path("/data/string/{data_key:path}/{data_file_id}")
    def data__string__with__id_and_key(self, cache_id     : Cache_Id             = None                          ,
                                             namespace    : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                             data_key     : Safe_Str__File__Path = None                          ,
                                             data_file_id : Safe_Str__Id         = None
                                       ) -> Response:                                                           # Retrieve data as plain text string
        request = Schema__Cache__Data__Retrieve__Request(cache_id     = cache_id                ,
                                                         data_type    = Enum__Cache__Data_Type.STRING,
                                                         data_key     = data_key                ,
                                                         data_file_id = data_file_id            ,
                                                         namespace    = namespace               )

        result = self.retrieve_service().retrieve_data(request)

        return self.handle_string_result(result)

    @route_path("/data/binary/{data_file_id}")
    def data__binary__with__id(self, cache_id     : Cache_Id             = None                          ,
                                     namespace    : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                     data_file_id : Safe_Str__Id         = None
                               ) -> Response:
        return self.data__binary__with__id_and_key(cache_id     = cache_id    ,
                                                   namespace    = namespace   ,
                                                   data_key     = ''          ,
                                                   data_file_id = data_file_id)

    @route_path("/data/binary/{data_key:path}/{data_file_id}")
    def data__binary__with__id_and_key(self, cache_id     : Cache_Id             = None                          ,
                                             namespace    : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                             data_key     : Safe_Str__File__Path = None                          ,
                                             data_file_id : Safe_Str__Id         = None
                                       ) -> Response:                                                           # Retrieve data as binary octet-stream
        request = Schema__Cache__Data__Retrieve__Request(cache_id     = cache_id                 ,
                                                         data_type    = Enum__Cache__Data_Type.BINARY,
                                                         data_key     = data_key                 ,
                                                         data_file_id = data_file_id             ,
                                                         namespace    = namespace                )

        result = self.retrieve_service().retrieve_data(request)

        return self.handle_binary_result(result)

    @type_safe
    def handle_not_found(self, result        : Union[Type_Safe, Dict] = None,                                  # Base method for 404 handling
                               cache_id      : Cache_Id               = None,
                               data_file_id  : Safe_Str__Id           = None):
        if result is None or (hasattr(result, 'found') and not result.found):
            error_detail = { "error_type"   : "NOT_FOUND"                                                     ,
                             "message"      : "Data file not found"                                            ,
                             "cache_id"     : str(cache_id) if cache_id else None                              ,
                             "data_file_id" : str(data_file_id) if data_file_id else None                      }
            raise HTTPException(status_code=404, detail=error_detail)
        return result

    def handle_json_result(self, result: Schema__Cache__Data__Retrieve__Response,                              # Handle JSON-specific result processing
                                 cache_id: Cache_Id    = None,
                                 data_file_id: Safe_Str__Id = None) -> dict:
        self.handle_not_found(result, cache_id=cache_id, data_file_id=data_file_id)
        return result.data

    def handle_string_result(self, result: Schema__Cache__Data__Retrieve__Response) -> Response:               # Handle string-specific result processing
        if not result or not result.found:
            return Response(content="", status_code=404)                                # return no content if not found

        return Response(content=result.data, media_type="text/plain")

    def handle_binary_result(self, result: Schema__Cache__Data__Retrieve__Response) -> Response:               # Handle binary-specific result processing
        if not result or not result.found:
            return Response(content=b"", status_code=404)                               # return no content if not found

        return Response(content=result.data, media_type="application/octet-stream")

    def setup_routes(self):                                                        # Configure all data retrieval routes
        self.add_route_get(self.data__json__with__id           )                                               # json endpoints
        self.add_route_get(self.data__json__with__id_and_key   )

        self.add_route_get(self.data__string__with__id         )                                                # string endpoints
        self.add_route_get(self.data__string__with__id_and_key )

        self.add_route_get(self.data__binary__with__id         )                                               # binary endpoints
        self.add_route_get(self.data__binary__with__id_and_key )


from osbot_fast_api.api.routes.Fast_API__Routes                     import Fast_API__Routes
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service    import DEFAULT_CACHE__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service           import Cache__Service
from mgraph_ai_service_cache.utils.testing.Cache__Test__Fixtures    import Cache__Test__Fixtures

TAG__ROUTES_SERVER                  = 'server'
ROUTES_PATHS__SERVER                = [f'/{TAG__ROUTES_SERVER}' + '/storage/info' ,
                                       f'/{TAG__ROUTES_SERVER}' + '/create/test-fixtures' ]

class Routes__Server(Fast_API__Routes):
    tag         : str          = TAG__ROUTES_SERVER
    cache_service : Cache__Service

    def storage__info(self) -> dict:                                                                     # Get current storage backend information
        return self.cache_service.get_storage_info()

    def create__test_fixtures(self):
        with Cache__Test__Fixtures(cache_service=self.cache_service     ,
                                   namespace  = DEFAULT_CACHE__NAMESPACE) as _:
            _.setup()
            return _.load_manifest__data()

    def setup_routes(self):
        self.add_route_get(self.storage__info        )
        self.add_route_get(self.create__test_fixtures)




import mgraph_ai_service_cache__web_console
from osbot_fast_api_serverless.deploy.Deploy__Serverless__Fast_API   import Deploy__Serverless__Fast_API
from mgraph_ai_service_cache.config                                  import SERVICE_NAME, LAMBDA_DEPENDENCIES__FAST_API_SERVERLESS
from mgraph_ai_service_cache.fast_api.lambda_handler                 import run

class Deploy__Service(Deploy__Serverless__Fast_API):


    def deploy_lambda(self):
        with super().deploy_lambda() as _:
            _.add_folder(mgraph_ai_service_cache__web_console.path)
            return _

    def handler(self):
        return run

    def lambda_dependencies(self):
        return LAMBDA_DEPENDENCIES__FAST_API_SERVERLESS

    def lambda_name(self):
        return SERVICE_NAME


from osbot_fast_api.api.routes.Fast_API__Routes                                                 import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                      import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                         import Safe_Str__Fast_API__Route__Tag
from mgraph_ai_service_cache_client.schemas.cache.file.Schema__Cache__Exists__Response          import Schema__Cache__Exists__Response
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                              import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                 import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash        import Safe_Str__Cache_Hash
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                              import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                       import Cache__Service

TAG__ROUTES_EXISTS                  = 'exists'
PREFIX__ROUTES_EXISTS               = '/{namespace}'
BASE_PATH__ROUTES_EXISTS            = f'{PREFIX__ROUTES_EXISTS}/{TAG__ROUTES_EXISTS}/'
ROUTES_PATHS__EXISTS                = [ BASE_PATH__ROUTES_EXISTS + 'hash/{cache_hash}',
                                        BASE_PATH__ROUTES_EXISTS + 'id/{cache_id}'    ]


class Routes__File__Exists(Fast_API__Routes):
    tag           : Safe_Str__Fast_API__Route__Tag     = TAG__ROUTES_EXISTS
    prefix        : Safe_Str__Fast_API__Route__Prefix  = PREFIX__ROUTES_EXISTS
    cache_service : Cache__Service


    def exists__id__cache_id(self, cache_id  : Cache_Id                             ,       # Check if cache_id exists
                                   namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                              ) -> Schema__Cache__Exists__Response:
        handler = self.cache_service.get_or_create_handler(namespace)
        file_id = Safe_Str__Id(cache_id)

        with handler.fs__refs_id.file__json__single(file_id) as ref_fs:
            exists = ref_fs.exists()

        return Schema__Cache__Exists__Response(exists    = exists   ,
                                               cache_id  = cache_id ,
                                               namespace = namespace)


    def exists__hash__cache_hash(self, cache_hash : Safe_Str__Cache_Hash                    ,   # Check if hash exists
                                       namespace  : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                ) -> Schema__Cache__Exists__Response:
        handler = self.cache_service.get_or_create_handler(namespace)
        file_id = Safe_Str__Id(cache_hash)

        with handler.fs__refs_hash.file__json__single(file_id) as ref_fs:
            exists = ref_fs.exists()

        return Schema__Cache__Exists__Response(exists     = exists    ,
                                               cache_hash = cache_hash,
                                               namespace  = namespace )

    def setup_routes(self):
        self.add_route_get(self.exists__id__cache_id    )
        self.add_route_get(self.exists__hash__cache_hash)


from typing                                                                                     import List, Optional
from osbot_utils.decorators.methods.cache_on_self                                               import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                            import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path               import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                              import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                 import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                  import type_safe
from osbot_utils.utils.Http                                                                     import url_join_safe
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__List__Response      import Schema__Cache__Data__List__Response
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__File__Info          import Schema__Cache__Data__File__Info
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                  import Enum__Cache__Data_Type
from mgraph_ai_service_cache.service.cache.Cache__Service                                       import Cache__Service
from mgraph_ai_service_cache.service.cache.retrieve.Cache__Service__Retrieve                    import Cache__Service__Retrieve


class Cache__Service__Data__List(Type_Safe):                                                    # Service for listing data files in cache entries
    cache_service : Cache__Service

    @cache_on_self
    def retrieve_service(self):
        return Cache__Service__Retrieve(cache_service=self.cache_service)

    def get_type_from_extension(self, extension: str) -> Enum__Cache__Data_Type:                # Map extension to data type
        ext_map = { 'json' : Enum__Cache__Data_Type.JSON  ,
                    'txt'  : Enum__Cache__Data_Type.STRING,
                    'bin'  : Enum__Cache__Data_Type.BINARY }
        return ext_map.get(extension, Enum__Cache__Data_Type.BINARY)

    @type_safe
    def list_data_files(self,
                        cache_id  : Cache_Id             ,
                        namespace : Safe_Str__Id         ,
                        data_key  : Safe_Str__File__Path = None,
                        recursive : bool                 = True
                   ) -> Optional[Schema__Cache__Data__List__Response]:                          # List all data files under a cache entry

        file_refs = self.retrieve_service().retrieve_by_id__refs(cache_id  = cache_id,
                                                                 namespace = namespace)

        if not file_refs:
            return None                                                                         # Cache entry doesn't exist

        data_folders = file_refs.file_paths.data_folders

        if not data_folders:
            return Schema__Cache__Data__List__Response(cache_id   = cache_id  ,                 # Return empty response if no data folders
                                                       namespace  = namespace ,
                                                       data_key   = data_key  ,
                                                       file_count = 0         ,
                                                       files      = []        ,
                                                       total_size = 0         )

        handler    = self.cache_service.get_or_create_handler(namespace)
        files      = []
        total_size = 0

        for data_folder in data_folders:
            target_folder = data_folder                                                         # Build target folder path

            if data_key:
                target_folder = url_join_safe(data_folder, data_key)

            file_paths = handler.storage_backend.folder__files__all(parent_folder=target_folder)  # Always use folder__files__all (only available method)

            if not recursive:                                                                   # Filter for non-recursive: only files directly in target_folder
                target_folder_str = str(target_folder)
                if not target_folder_str.endswith('/'):
                    target_folder_str += '/'
                file_paths = [fp for fp in file_paths
                              if '/' not in str(fp).replace(target_folder_str, '', 1)]          # No subdirectories

            for file_path in file_paths:                                                        # Parse file info for each file
                if file_path.endswith('.config') or file_path.endswith('.metadata'):            # Skip config and metadata files
                    continue

                file_info = self._parse_file_info(file_path       = file_path      ,
                                                  data_folder     = data_folder    ,
                                                  storage_backend = handler.storage_backend)
                if file_info:
                    files.append(file_info)
                    if file_info.file_size:
                        total_size += file_info.file_size

        return Schema__Cache__Data__List__Response(cache_id   = cache_id      ,
                                                   namespace  = namespace     ,
                                                   data_key   = data_key      ,
                                                   file_count = len(files)    ,
                                                   files      = files         ,
                                                   total_size = total_size    )

    def _parse_file_info(self,
                         file_path       : str,
                         data_folder     : str,
                         storage_backend
                    ) -> Schema__Cache__Data__File__Info:                                    # Parse a file path into Schema__Cache__Data__File__Info

        try:
            filename = file_path.split('/')[-1]                                                 # Extract filename from path

            if '.' in filename:                                                                 # Parse extension and data_file_id
                parts        = filename.rsplit('.', 1)
                data_file_id = parts[0]
                extension    = parts[1]
            else:
                data_file_id = filename
                extension    = ''

            data_type = self.get_type_from_extension(extension)                                 # Determine data type from extension

            data_folder_with_slash = data_folder if data_folder.endswith('/') else data_folder + '/'
            relative_path = file_path.replace(data_folder_with_slash, '', 1)                    # Extract data_key from path

            if '/' in relative_path:
                data_key = '/'.join(relative_path.split('/')[:-1])
            else:
                data_key = ''

            file_size = self._get_file_size(storage_backend, file_path)                         # Get file size safely

            return Schema__Cache__Data__File__Info(data_file_id = data_file_id,
                                                   data_key     = data_key    ,
                                                   data_type    = data_type   ,
                                                   file_path    = file_path   ,
                                                   file_size    = file_size   ,
                                                   extension    = extension   )
        except Exception:
            return None

    def _get_file_size(self, storage_backend, file_path: str) -> int:                           # Get file size safely - not all backends have file__size
        try:
            if hasattr(storage_backend, 'file__size'):
                return storage_backend.file__size(file_path) or 0
            file_bytes = storage_backend.file__bytes(file_path)                                 # Fallback: read bytes and get length
            return len(file_bytes) if file_bytes else 0
        except Exception:
            return 0


from typing                                                                                 import List, Dict, Any
from osbot_utils.type_safe.Type_Safe                                                        import Type_Safe
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text                import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash    import Safe_Str__Cache_Hash
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                          import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id             import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now            import Timestamp_Now
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type              import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy        import Enum__Cache__Store__Strategy


class Schema__Test_Data__Entry(Type_Safe):                                       # Schema for individual created cache entry
    cache_id    : Cache_Id
    cache_hash  : Safe_Str__Cache_Hash
    namespace   : Safe_Str__Id
    strategy    : Enum__Cache__Store__Strategy
    data_type   : Enum__Cache__Data_Type
    description : Safe_Str__Text
    cache_key   : Safe_Str__File__Path
    children    : List[Dict[str, Any]]                                           # Optional: for entries with child data


class Schema__Test_Data__Create__Response(Type_Safe):                            # Response schema for test data creation
    success         : bool
    timestamp       : Timestamp_Now
    entries_created : int
    namespaces      : List[Safe_Str__Id]
    strategies_used : List[str]
    entries         : List[Dict[str, Any]]
    message         : Safe_Str__Text


class Schema__Test_Data__Clear__Response(Type_Safe):                             # Response schema for clearing namespace
    success       : bool
    namespace     : Safe_Str__Id
    files_deleted : int
    message       : Safe_Str__Text
    error         : Safe_Str__Text                                                 # Only populated on failure


class Schema__Test_Data__Sample_Json(Type_Safe):                                 # Schema for sample JSON test data
    type        : Safe_Str__Id
    description : Safe_Str__Text


class Schema__Test_Data__Key_Based_Item(Type_Safe):                              # Schema for key-based test data item
    cache_key : Safe_Str__Id
    file_id   : Safe_Str__Id
    data      : Dict[str, Any]


class Schema__Test_Data__Child_Item(Type_Safe):                                  # Schema for child data item
    data_key     : Safe_Str__Id
    data_file_id : Safe_Str__Id
    data         : Dict[str, Any]


class Schema__Test_Data__Zip_File(Type_Safe):                                    # Schema for ZIP file content
    path    : Safe_Str__Id
    content : bytes


from typing                                                                                          import List, Dict, Any
from osbot_utils.type_safe.Type_Safe                                                                 import Type_Safe
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                      import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now                     import Timestamp_Now
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                    import Safe_Str__File__Path
from osbot_utils.decorators.methods.cache_on_self                                                    import cache_on_self
from osbot_utils.utils.Zip                                                                           import zip_bytes_empty
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy                 import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                       import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Store__Request           import Schema__Cache__Data__Store__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Store__Request             import Schema__Cache__Zip__Store__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Operation__Request         import Schema__Cache__Zip__Operation__Request
from mgraph_ai_service_cache.service.cache.Cache__Service                                            import Cache__Service
from mgraph_ai_service_cache.service.cache.store.Cache__Service__Store                               import Cache__Service__Store
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__Store                          import Cache__Service__Data__Store
from mgraph_ai_service_cache.service.cache.zip.Cache__Service__Zip__Store                            import Cache__Service__Zip__Store
from mgraph_ai_service_cache.service.cache.zip.Cache__Service__Zip__Operations                       import Cache__Service__Zip__Operations
from mgraph_ai_service_cache.schemas.test_data.Schema__Test_Data                                     import Schema__Test_Data__Create__Response
from mgraph_ai_service_cache.schemas.test_data.Schema__Test_Data                                     import Schema__Test_Data__Clear__Response

DEFAULT_NAMESPACES = [ 'test-data', 'test-demo', 'test-analytics'] # 'default',
DEFAULT_STRATEGIES = [ Enum__Cache__Store__Strategy.DIRECT          ,
                       Enum__Cache__Store__Strategy.TEMPORAL        ,
                       Enum__Cache__Store__Strategy.TEMPORAL_LATEST ,
                       Enum__Cache__Store__Strategy.KEY_BASED       ]


class Cache__Service__Test_Data(Type_Safe):                                      # Service for generating test data for the web console
    cache_service : Cache__Service

    @cache_on_self
    def store_service(self) -> Cache__Service__Store:
        return Cache__Service__Store(cache_service=self.cache_service)

    @cache_on_self
    def data_store_service(self) -> Cache__Service__Data__Store:
        return Cache__Service__Data__Store(cache_service=self.cache_service)

    @cache_on_self
    def zip_store_service(self) -> Cache__Service__Zip__Store:
        return Cache__Service__Zip__Store(cache_service=self.cache_service)

    @cache_on_self
    def zip_ops_service(self) -> Cache__Service__Zip__Operations:
        return Cache__Service__Zip__Operations(cache_service=self.cache_service)

    def create_comprehensive(self) -> Schema__Test_Data__Create__Response:       # Create comprehensive test data across all strategies and namespaces
        entries    = []
        namespaces = DEFAULT_NAMESPACES
        strategies = DEFAULT_STRATEGIES

        for namespace in namespaces:
            namespace_entries = self._create_namespace_test_data(namespace, strategies)
            entries.extend(namespace_entries)

        return Schema__Test_Data__Create__Response(success         = True                             ,
                                                   timestamp       = Timestamp_Now()                  ,
                                                   entries_created = len(entries)                     ,
                                                   namespaces      = namespaces                       ,
                                                   strategies_used = [s.value for s in strategies]    ,
                                                   entries         = entries                          ,
                                                   message         = 'Comprehensive test data created')

    def create_for_namespace(self, namespace: Safe_Str__Id                       # Create test data for a specific namespace
                             ) -> Schema__Test_Data__Create__Response:
        strategies = DEFAULT_STRATEGIES
        entries    = self._create_namespace_test_data(str(namespace), strategies)

        return Schema__Test_Data__Create__Response(success         = True                                ,
                                                   timestamp       = Timestamp_Now()                     ,
                                                   entries_created = len(entries)                        ,
                                                   namespaces      = [str(namespace)]                    ,
                                                   strategies_used = [s.value for s in strategies]       ,
                                                   entries         = entries                             ,
                                                   message         = f'Test data created for {namespace}')

    def create_minimal(self) -> Schema__Test_Data__Create__Response:             # Create minimal test data (quick setup)
        namespace = 'default'
        entries   = self._create_json_entries(namespace, Enum__Cache__Store__Strategy.DIRECT, count=3)

        return Schema__Test_Data__Create__Response(success         = True                        ,
                                                   timestamp       = Timestamp_Now()             ,
                                                   entries_created = len(entries)                ,
                                                   namespaces      = [namespace]                 ,
                                                   strategies_used = ['direct']                  ,
                                                   entries         = entries                     ,
                                                   message         = 'Minimal test data created')

    def clear_namespace(self, namespace: Safe_Str__Id                            # Clear all data from a specific namespace
                        ) -> Schema__Test_Data__Clear__Response:
        try:
            storage_fs = self.store_service().cache_service.storage_fs()
            files      = storage_fs.folder__files__all(parent_folder=namespace)
            deleted    = 0

            for file_path in files:
                if storage_fs.file__delete(file_path):
                    deleted += 1

            return Schema__Test_Data__Clear__Response(success       = True                                              ,
                                                      namespace     = namespace                                         ,
                                                      files_deleted = deleted                                           ,
                                                      message       = f'Cleared {deleted} files from {namespace}',
                                                      error         = ''                                        )
        except Exception as e:
            return Schema__Test_Data__Clear__Response(success       = False     ,
                                                      namespace     = namespace ,
                                                      files_deleted = 0         ,
                                                      message       = ''        ,
                                                      error         = str(e)    )

    def _create_namespace_test_data(self, namespace: str,                        # Create all test data for a single namespace
                                    strategies: List[Enum__Cache__Store__Strategy]
                                    ) -> List[Dict]:
        entries = []

        for strategy in strategies:
            json_entries = self._create_json_entries(namespace, strategy)
            entries.extend(json_entries)

            string_entries = self._create_string_entries(namespace, strategy)
            entries.extend(string_entries)

            if strategy == Enum__Cache__Store__Strategy.KEY_BASED:
                key_entries = self._create_key_based_entries(namespace)
                entries.extend(key_entries)

        if namespace == 'default':                                               # Special data for default namespace
            child_entries = self._create_entries_with_child_data(namespace)
            entries.extend(child_entries)

            zip_entries = self._create_zip_entries(namespace)
            entries.extend(zip_entries)

        return entries

    def _create_json_entries(self, namespace: str,                               # Create JSON cache entries
                             strategy: Enum__Cache__Store__Strategy,
                             count: int = 5
                             ) -> List[Dict]:
        entries   = []
        test_data = self._get_sample_json_data()

        for i, data in enumerate(test_data[:count]):
            result = self.store_service().store_json(data      = data      ,
                                                     strategy  = strategy  ,
                                                     namespace = namespace)
            if result:
                entries.append({ "cache_id"    : str(result.cache_id)                      ,
                                 "cache_hash"  : str(result.cache_hash)                    ,
                                 "namespace"   : namespace                                 ,
                                 "strategy"    : strategy.value                            ,
                                 "data_type"   : Enum__Cache__Data_Type.JSON.value         ,
                                 "description" : data.get('description', f'JSON entry {i}')})
        return entries

    def _create_string_entries(self, namespace: str,                             # Create string cache entries
                               strategy: Enum__Cache__Store__Strategy,
                               count: int = 3
                               ) -> List[Dict]:
        entries      = []
        test_strings = [ "Simple text content for cache testing"                                          ,
                         "Configuration: debug=true, log_level=INFO, max_connections=100"                 ,
                         "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed do eiusmod tempor."]

        for i, data in enumerate(test_strings[:count]):
            result = self.store_service().store_string(data      = data      ,
                                                       strategy  = strategy  ,
                                                       namespace = namespace)
            if result:
                entries.append({ "cache_id"    : str(result.cache_id)                ,
                                 "cache_hash"  : str(result.cache_hash)              ,
                                 "namespace"   : namespace                           ,
                                 "strategy"    : strategy.value                      ,
                                 "data_type"   : Enum__Cache__Data_Type.STRING.value ,
                                 "description" : f'String entry {i}'                 })
        return entries

    def _create_key_based_entries(self, namespace: str) -> List[Dict]:           # Create key-based organized entries
        entries        = []
        key_based_data = self._get_key_based_test_data()

        for item in key_based_data:
            result = self.store_service().store_json(data      = item["data"]                             ,
                                                     strategy  = Enum__Cache__Store__Strategy.KEY_BASED   ,
                                                     namespace = namespace                                ,
                                                     cache_key = item["cache_key"]                        ,
                                                     file_id   = item["file_id"]                          )
            if result:
                entries.append({ "cache_id"    : str(result.cache_id)              ,
                                 "cache_hash"  : str(result.cache_hash)            ,
                                 "namespace"   : namespace                         ,
                                 "strategy"    : "key_based"                       ,
                                 "data_type"   : Enum__Cache__Data_Type.JSON.value ,
                                 "cache_key"   : item["cache_key"]                 ,
                                 "description" : f'Key-based: {item["cache_key"]}' })
        return entries

    def _create_entries_with_child_data(self, namespace: str) -> List[Dict]:     # Create parent entries with child data files
        entries     = []
        parent_data = { "type"        : "document"                          ,
                        "title"       : "Analysis Report"                   ,
                        "created"     : "2025-01-09T10:30:00Z"               ,
                        "author"      : "system"                            ,
                        "description" : "Document with child data files"    }

        result = self.store_service().store_json(data      = parent_data                                 ,
                                                 strategy  = Enum__Cache__Store__Strategy.TEMPORAL_LATEST,
                                                 namespace = Safe_Str__Id(namespace)                     )

        if result:
            parent_entry = { "cache_id"    : str(result.cache_id)              ,
                             "cache_hash"  : str(result.cache_hash)            ,
                             "namespace"   : namespace                         ,
                             "strategy"    : "temporal_latest"                 ,
                             "data_type"   : Enum__Cache__Data_Type.JSON.value ,
                             "description" : "Parent with child data"          ,
                             "children"    : []                                }

            child_data_items = self._get_child_data_items()

            for child in child_data_items:
                request = Schema__Cache__Data__Store__Request(cache_id     = result.cache_id                              ,
                                                              data         = child["data"]                                ,
                                                              data_type    = Enum__Cache__Data_Type.JSON                  ,
                                                              data_key     = Safe_Str__File__Path(child["data_key"])      ,
                                                              data_file_id = Safe_Str__Id(child["data_file_id"])          ,
                                                              namespace    = Safe_Str__Id(namespace)                      )
                child_result = self.data_store_service().store_data(request)
                if child_result:
                    parent_entry["children"].append({ "data_key"     : child["data_key"]     ,
                                                      "data_file_id" : child["data_file_id"] })

            entries.append(parent_entry)

        return entries

    def _create_zip_entries(self, namespace: str) -> List[Dict]:                 # Create ZIP package entries
        entries   = []
        zip_bytes = zip_bytes_empty()

        request    = Schema__Cache__Zip__Store__Request(zip_bytes = zip_bytes                                     ,
                                                        cache_key = Safe_Str__File__Path("packages/sample-app")   ,
                                                        file_id   = Safe_Str__Id("v1.0.0")                        ,
                                                        namespace = Safe_Str__Id(namespace)                       ,
                                                        strategy  = Enum__Cache__Store__Strategy.KEY_BASED        )
        zip_result = self.zip_store_service().store_zip(request)

        if zip_result and zip_result.success:
            files_to_add     = self._get_zip_file_contents()
            current_cache_id = zip_result.cache_id

            for file_info in files_to_add:
                add_request = Schema__Cache__Zip__Operation__Request(cache_id     = current_cache_id                         ,
                                                                     operation    = "add"                                    ,
                                                                     file_path    = Safe_Str__File__Path(file_info["path"])  ,
                                                                     file_content = file_info["content"]                     ,
                                                                     namespace    = Safe_Str__Id(namespace)                  )
                add_result = self.zip_ops_service().perform_operation(add_request)
                if add_result and add_result.success and add_result.cache_id:
                    current_cache_id = add_result.cache_id

            entries.append({ "cache_id"    : str(current_cache_id)           ,
                             "cache_hash"  : str(zip_result.cache_hash)      ,
                             "namespace"   : namespace                       ,
                             "strategy"    : "key_based"                     ,
                             "data_type"   : "zip"                           ,
                             "cache_key"   : "packages/sample-app"           ,
                             "description" : "Sample ZIP package with files" })

        return entries

    def _get_sample_json_data(self) -> List[Dict]:                               # Sample JSON data for testing
        return [
            { "type"        : "user_profile"                                      ,
              "user_id"     : "usr_12345"                                         ,
              "email"       : "test@example.com"                                  ,
              "preferences" : { "theme": "dark", "notifications": True }          ,
              "description" : "Sample user profile"                               },

            { "type"        : "api_response"                                      ,
              "endpoint"    : "/api/v1/data"                                      ,
              "status"      : 200                                                 ,
              "data"        : { "items": [1, 2, 3], "total": 3 }                  ,
              "description" : "Cached API response"                               },

            { "type"        : "analytics_event"                                   ,
              "event_name"  : "page_view"                                         ,
              "timestamp"   : "2025-01-09T14:30:00Z"                              ,
              "properties"  : { "page": "/dashboard", "duration_ms": 1500 }       ,
              "description" : "Analytics tracking event"                          },

            { "type"        : "configuration"                                     ,
              "service"     : "cache-service"                                     ,
              "version"     : "0.5.68"                                            ,
              "settings"    : { "max_size": 1024, "ttl": 3600, "compression": True},
              "description" : "Service configuration"                             },

            { "type"        : "workflow_state"                                    ,
              "workflow_id" : "wf_abc123"                                         ,
              "status"      : "completed"                                         ,
              "steps"       : ["init", "process", "validate", "complete"]         ,
              "description" : "Workflow execution state"                          },

            { "type"        : "model_output"                                      ,
              "model"       : "sentiment-analyzer"                                ,
              "input_hash"  : "abc123def456"                                      ,
              "output"      : { "sentiment": "positive", "confidence": 0.92 }     ,
              "description" : "ML model cached output"                            },

            { "type"        : "session_data"                                      ,
              "session_id"  : "sess_xyz789"                                       ,
              "user_id"     : "usr_12345"                                         ,
              "expires_at"  : "2025-01-10T14:30:00Z"                              ,
              "description" : "User session information"                          }
        ]

    def _get_key_based_test_data(self) -> List[Dict]:                            # Key-based hierarchical test data
        return [
            { "cache_key" : "reports/2025/q1/sales"      ,
              "file_id"   : "quarterly-report"           ,
              "data"      : { "quarter"  : "Q1 2025"      ,
                              "revenue"  : 1250000       ,
                              "growth"   : 12.5          ,
                              "regions"  : ["NA", "EMEA"]}},

            { "cache_key" : "reports/2025/q1/users"      ,
              "file_id"   : "user-metrics"               ,
              "data"      : { "quarter"   : "Q1 2025"     ,
                              "new_users" : 15420        ,
                              "active"    : 89340        ,
                              "churn"     : 2.1          }},

            { "cache_key" : "config/services/cache"      ,
              "file_id"   : "cache-config"               ,
              "data"      : { "ttl_hours"    : 24         ,
                              "max_size_mb"  : 1024       ,
                              "compression"  : True       ,
                              "storage_mode" : "s3"       }},

            { "cache_key" : "config/services/api"        ,
              "file_id"   : "api-config"                 ,
              "data"      : { "rate_limit"    : 1000       ,
                              "timeout_ms"    : 30000      ,
                              "retry_count"   : 3          ,
                              "circuit_break" : True       }},

            { "cache_key" : "logs/2025/01/09"            ,
              "file_id"   : "daily-summary"              ,
              "data"      : { "date"           : "2025-01-09",
                              "requests"       : 156780     ,
                              "errors"         : 234        ,
                              "avg_latency_ms" : 45         }}
        ]

    def _get_child_data_items(self) -> List[Dict]:                               # Child data items for parent entries
        return [
            { "data_key"     : "analysis/sentiment"                        ,
              "data_file_id" : "sentiment-v1"                              ,
              "data"         : { "positive" : 0.72                         ,
                                 "negative" : 0.18                         ,
                                 "neutral"  : 0.10                         }},

            { "data_key"     : "analysis/keywords"                         ,
              "data_file_id" : "keywords-v1"                               ,
              "data"         : { "keywords" : ["cache", "performance", "api"],
                                 "scores"   : [0.95, 0.87, 0.82]           }},

            { "data_key"     : "metadata"                                  ,
              "data_file_id" : "file-metadata"                             ,
              "data"         : { "source"    : "import"                    ,
                                 "processed" : True                        ,
                                 "version"   : 1                           }}
        ]

    def _get_zip_file_contents(self) -> List[Dict]:                              # Files to add to ZIP package
        return [
            { "path"    : "README.md"                                                 ,
              "content" : b"# Sample Application\n\nThis is a test package."         },

            { "path"    : "config.json"                                              ,
              "content" : b'{"name": "sample-app", "version": "1.0.0"}'               },

            { "path"    : "src/main.py"                                              ,
              "content" : b'print("Hello from cache!")'                              },

            { "path"    : "src/utils.py"                                             ,
              "content" : b'def helper(): return True'                               }
        ]


import gzip
import json
from typing                                                                                      import Dict, Optional, Any, List, Union
from mgraph_ai_service_cache_client.schemas.cache.file.Schema__Cache__File__Metadata             import Schema__Cache__File__Metadata
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__File__Cache_Hash     import Safe_Str__Cache__File__Cache_Hash
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__File__Cache_Key      import Safe_Str__Cache__File__Cache_Key
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__File__File_Id        import Safe_Str__Cache__File__File_Id
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__Namespace            import Safe_Str__Cache__Namespace
from osbot_utils.helpers.cache.Cache__Hash__Generator                                            import Cache__Hash__Generator
from osbot_utils.helpers.cache.schemas.Schema__Cache__Hash__Config                               import Schema__Cache__Hash__Config
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid                            import Random_Guid
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                   import type_safe
from memory_fs.schemas.Schema__Memory_FS__File__Config                                           import Schema__Memory_FS__File__Config
from osbot_utils.decorators.methods.cache_on_self                                                import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                             import Type_Safe
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                  import Safe_Str__Id
from osbot_utils.utils.Files                                                                     import file_extension, file_name_without_extension
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                               import Cache_Id
from osbot_utils.utils.Http                                                                      import url_join_safe
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash         import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.service.cache_service.Schema__Store__Context                import Schema__Store__Context       # todo: review this schema since it is not currently stored in the client project
from mgraph_ai_service_cache_client.schemas.cache.file.Schema__Cache__File__Refs                 import Schema__Cache__File__Refs
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service                          import DEFAULT_CACHE__NAMESPACE
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy             import Enum__Cache__Store__Strategy
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Json__Field_Path    import Safe_Str__Json__Field_Path
from mgraph_ai_service_cache.service.cache.Cache__Config                                         import Cache__Config
from mgraph_ai_service_cache.service.cache.Cache__Handler                                        import Cache__Handler
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Store__Response                 import Schema__Cache__Store__Response
from mgraph_ai_service_cache.service.cache.store.Cache__Service__Store__With_Strategy            import Cache__Service__Store__With_Strategy


class Cache__Service(Type_Safe):                                                    # Main cache service orchestrator
    cache_config      : Cache__Config                                               # Configuration object
    cache_handlers    : Dict[Safe_Str__Id, Cache__Handler]                          # Multiple cache handlers by namespace
    hash_config       : Schema__Cache__Hash__Config                                         # Hash generation config
    hash_generator    : Cache__Hash__Generator                                      # Hash generator instance

    # todo: I think we can now remove this file and just use the self.storage_fs()
    @cache_on_self
    def storage_backend(self):                                                      # Create and cache the storage backend
        return self.cache_config.create_storage_backend()

    @cache_on_self
    def storage_fs(self):                                                           # Return storage backend for direct operations
        return self.storage_backend()                                               # This is used for admin operations that need direct storage access

    # todo: this logic is starting to be quite complex to be in a method, I think we can refactor this logic into a separate class and have methods for
    #       each logic step/action
    # todo: refactor to add type safe return type
    def delete_by_id(self, cache_id: Cache_Id, namespace: Safe_Str__Id = None) -> Dict[str, Any]:
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)

        with handler.fs__refs_id.file__json__single(Safe_Str__Id(str(cache_id))) as ref_fs:
            if not ref_fs.exists():
                return {"status": "not_found", "message": f"Cache ID {cache_id} not found"}

            id_ref_data  = ref_fs.content()
            all_paths    = id_ref_data.get("all_paths", {})
            cache_hash   = id_ref_data.get("cache_hash")
            strategy     = id_ref_data.get("strategy")

        deleted_paths = []                                                                          # Track deletion results
        failed_paths  = []

        fs_data = handler.get_fs_for_strategy(strategy)                                             # Delete data files first (use the appropriate fs based on strategy)
        for path in all_paths.get("data", []):
            try:
                if fs_data.storage_fs.file__delete(path):
                    deleted_paths.append(path)
                else:
                    failed_paths.append(path)
            except Exception as e:
                failed_paths.append(f"{path}: {str(e)}")

        if cache_hash:                                                                              # Update hash reference (remove this cache_id from the list)
            with handler.fs__refs_hash.file__json__single(Safe_Str__Id(cache_hash)) as ref_fs:
                if ref_fs.exists():
                    refs                   = ref_fs.content()
                    refs["cache_ids"]      = [entry for entry in refs["cache_ids"]                  # Remove this cache_id from the list
                                              if entry["cache_id"] != str(cache_id)]
                    refs["total_versions"] -= 1

                    if refs["total_versions"] > 0:
                        if refs["latest_id"] == str(cache_id) and refs["cache_ids"]:                # Update the latest_id if needed
                            refs["latest_id"] = refs["cache_ids"][-1]["cache_id"]
                        ref_fs.update(file_data=refs)
                    else:
                        for path in all_paths.get("by_hash", []):                                   # No more versions, delete the hash reference files
                            try:
                                if handler.fs__refs_hash.storage_fs.file__delete(path):
                                    deleted_paths.append(path)
                                else:
                                    failed_paths.append(path)
                            except Exception as e:
                                failed_paths.append(f"{path}: {str(e)}")

        for path in all_paths.get("by_id", []):                                                 # Finally, delete the ID reference files
            try:
                if handler.fs__refs_id.storage_fs.file__delete(path):
                    deleted_paths.append(path)
                else:
                    failed_paths.append(path)
            except Exception as e:
                failed_paths.append(f"{path}: {str(e)}")

        return { "status"        : "success" if not failed_paths else "partial",                    # todo: this should be a Type_Safe class
                 "cache_id"      : str(cache_id)        ,
                 "deleted_count" : len(deleted_paths)   ,
                 "failed_count"  : len(failed_paths)    ,
                 "deleted_paths" : deleted_paths        ,
                 "failed_paths"  : failed_paths         }

    def get_all_namespaces_stats(self) -> Dict[str, Any]:                          # Get file counts for all active namespaces
        all_stats = {}

        for namespace in self.cache_handlers.keys():
            counts_data = self.get_namespace__file_counts(namespace)
            all_stats[str(namespace)] = {
                'total_files': counts_data['total_files'],
                'file_counts': counts_data['file_counts']
            }

        return { 'namespaces'       : all_stats                 ,                               # todo: this should be a Type_Safe class
                 'total_namespaces' : len(self.cache_handlers)  ,
                 'grand_total_files': sum(ns['total_files'] for ns in all_stats.values()),
                 'storage_mode'     : self.cache_config.storage_mode.value                }     # Include storage mode in stats

    # todo: BUG: rename form file_hashes to cache_hashes
    def get_namespace__file_hashes(self, namespace: Safe_Str__Cache__Namespace) -> List[str]:
        file_hashes = []
        handler = self.get_or_create_handler(namespace)

        parent_folder = url_join_safe(str(namespace), "refs/by-hash")                                           # Build the namespace-specific folder path

        for file_path in handler.fs__refs_hash.storage_fs.folder__files__all(parent_folder=parent_folder):      # Use folder__files__all to get only files in this namespace's refs/by-hash folder
            if file_extension(file_path) == '.json':
                file_id     = file_name_without_extension(file_path)                                            # Extract just the hash part (last segment after any sharding)
                hash_parts  = file_id.split('/')
                if hash_parts:
                    file_hashes.append(hash_parts[-1])
        return sorted(file_hashes)

    # todo: BUG: rename form file_hashes to cache_ids_hashes
    def get_namespace__file_ids(self, namespace: Safe_Str__Id) -> List[str]:
        file_ids = []
        handler = self.get_or_create_handler(namespace)

        parent_folder = url_join_safe(str(namespace), "refs/by-id")                                             # Build the namespace-specific folder path

        for file_path in handler.fs__refs_id.storage_fs.folder__files__all(parent_folder=parent_folder):        # Use folder__files__all to get only files in this namespace's refs/by-id folder
            if file_extension(file_path) == '.json':
                file_id = file_name_without_extension(file_path)
                id_parts = file_id.split('/')                                                                   # Extract just the ID part (last segment after any sharding)
                if id_parts:
                    file_ids.append(id_parts[-1])
        return sorted(file_ids)

    def get_namespace__file_counts(self, namespace: Safe_Str__Id = None) -> Dict[str, Any]:       # Get file counts for all strategies in a namespace
        namespace = namespace or Safe_Str__Id("default")
        handler = self.get_or_create_handler(namespace)

        file_counts = {}
        total_files = 0

        for strategy in ["direct", "temporal", "temporal_latest", "temporal_versioned"]:
            try:
                fs = handler.get_fs_for_strategy(strategy)
                if fs and fs.storage_fs:
                    count = len(fs.storage_fs.files__paths())
                    file_counts[f"{strategy}_files"] = count
                    total_files += count
                else:
                    file_counts[f"{strategy}_files"] = 0
            except Exception:
                file_counts[f"{strategy}_files"] = 0

        # Count reference store files
        try:
            refs_hash_count = len(handler.fs__refs_hash.storage_fs.files__paths())
            file_counts['refs_hash_files'] = refs_hash_count
            total_files += refs_hash_count
        except Exception:
            file_counts['refs_hash_files'] = 0

        try:
            refs_id_count = len(handler.fs__refs_id.storage_fs.files__paths())
            file_counts['refs_id_files'] = refs_id_count
            total_files += refs_id_count
        except Exception:
            file_counts['refs_id_files'] = 0

        file_counts['total_files'] = total_files

        return {
            'namespace': str(namespace),
            'handler': handler,
            'file_counts': file_counts,
            'total_files': total_files,
            'storage_mode': self.cache_config.storage_mode.value                   # Include storage mode
        }

    def get_or_create_handler(self, namespace: Safe_Str__Id = DEFAULT_CACHE__NAMESPACE) -> Cache__Handler:
        if namespace not in self.cache_handlers:                                                         # Create handler with shared storage backend and namespace
            handler = Cache__Handler(storage_backend  = self.storage_backend(),                          # Shared storage backend
                                     namespace        = str(namespace),                                  # Namespace for path prefixing
                                     cache_ttl_hours  = self.cache_config.default_ttl_hours).setup()
            self.cache_handlers[namespace] = handler
        return self.cache_handlers[namespace]

    def get_storage_info(self) -> Dict[str, Any]:                                  # Get information about current storage configuration
        return self.cache_config.get_storage_info()

    def hash__from_string(self, value) ->Safe_Str__Cache__File__Cache_Hash:
        cache_hash = self.hash_generator.from_string(value)
        return Safe_Str__Cache__File__Cache_Hash(cache_hash)                                  # do this cast because the hash_generator returns an object of type Safe_Str__Cache_Hash

    # todo: see if we can't use the Schema__Store__Context as the main param here
    @type_safe
    def store_with_strategy(self, storage_data     : Union[str, Dict, bytes]                              ,
                                  cache_hash       : Safe_Str__Cache_Hash                                 ,
                                  strategy         : Enum__Cache__Store__Strategy                         ,
                                  cache_id         : Cache_Id                           = None            ,
                                  cache_key        : Safe_Str__Cache__File__Cache_Key   = None                    ,
                                  file_id          : Safe_Str__Cache__File__File_Id     = None                    ,
                                  json_field_path  : Safe_Str__Json__Field_Path         = None                    ,
                                  namespace        : Safe_Str__Cache__Namespace         = DEFAULT_CACHE__NAMESPACE,
                                  content_encoding : Safe_Str__Id                       = None                    ,
                                  metadata         : Dict[str, Any]                     = None
                            ) -> Schema__Cache__Store__Response:                    # Store data using the specified strategy

        if not cache_hash:
            raise ValueError("in Cache__Service.store_with_strategy, the cache_hash must be provided")          # Validate inputs

        # todo this logic can be refactored to store_strategy.execute
        cache_id  = cache_id or Cache_Id(Random_Guid())
        cache_key = Safe_Str__Cache__File__Cache_Key(cache_key if cache_key else None)
        file_id   = Safe_Str__Cache__File__File_Id  (file_id or cache_id or Random_Guid())
        handler   = self.get_or_create_handler(namespace)
        context   = Schema__Store__Context(storage_data     = storage_data     ,                                  # Build context with all parameters
                                           cache_hash       = cache_hash       ,
                                           cache_id         = cache_id         ,
                                           cache_key        = cache_key        ,
                                           file_id          = file_id          ,
                                           json_field_path  = json_field_path  ,
                                           namespace        = namespace        ,
                                           strategy         = strategy         ,
                                           content_encoding = content_encoding ,
                                           handler          = handler          ,
                                           metadata         = metadata         )

        store_strategy = Cache__Service__Store__With_Strategy()
        return store_strategy.execute(context)                                                  # Execute storage strategy

    # todo: change return to type_safe value
    @type_safe
    def retrieve_by_hash(self, cache_hash : Safe_Str__Cache_Hash,
                               namespace  : Safe_Str__Id = None
                          ) -> Optional[Dict[str, Any]]:                        # Retrieve latest by hash"""
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)

        with handler.fs__refs_hash.file__json__single(Safe_Str__Id(cache_hash)) as ref_fs:   # Get hash->ID mapping
            if not ref_fs.exists():
                return None
            refs = ref_fs.content()
            latest_id = refs.get("latest_id")

        if not latest_id:
            return None

        return self.retrieve_by_id(Cache_Id(latest_id), namespace)           # Delegate to retrieve_by_id which handles the path lookup

    # todo: change return to type_safe value
    @type_safe
    def retrieve_by_hash__refs_hash(self, cache_hash : Safe_Str__Cache_Hash,
                               namespace  : Safe_Str__Id = None
                          ) -> Optional[Dict[str, Any]]:                        # Retrieve latest by hash"""
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)
        file_id   = Safe_Str__Id(cache_hash)                                    # Get hash->ID mapping
        with handler.fs__refs_hash.file__json__single(file_id=file_id) as ref_fs:
            if not ref_fs.exists():
                return None
            refs_hash = ref_fs.content()
            return refs_hash

    # todo: same as with the delete method above,  this logic is starting to be too complex to be all in one method
    #       also I think we are doing too much here, with far too many calls to the file system
    #       at least we should just return the data (i.e. we don't need to return the metadata here (since there is an enpoint to get that)
    def retrieve_by_id(self, cache_id  : Cache_Id,
                             namespace : Safe_Str__Id = DEFAULT_CACHE__NAMESPACE
                        ) -> Optional[Dict[str, Any]]:                     # todo: review this return value, since we had some exceptions here
        handler   = self.get_or_create_handler(namespace)

        # Get ID reference with content path
        with handler.fs__refs_id.file__json__single(Safe_Str__Id(cache_id)) as ref_fs:
            if not ref_fs.exists():
                return None
            ref_data = ref_fs.content()

        paths__content = ref_data.get("file_paths", {}).get("content_files")
        file_type     = ref_data.get("file_type", "json")

        if not paths__content:
            return None

        # Get the storage backend
        storage = handler.fs__refs_id.storage_fs

        # Read the content file directly (first path is the main content)
        content_path = paths__content[0] if paths__content else None

        if content_path and storage.file__exists(content_path):
            if file_type == "binary":
                data = storage.file__bytes(content_path)
            else:
                data = storage.file__json(content_path)

            # Read metadata                                         # todo: review the use of metadata here, since this probably one extra call we don't need to make (since we already have the data from the by-refs file
            metadata_path = content_path + '.metadata'              # todo: review this usage since we should have a much better way to do this using Memory_FS
            metadata_data = {}

            if storage.file__exists(metadata_path):
                metadata_raw = storage.file__json(metadata_path)
                # Convert metadata keys to Safe_Str__Id for consistency
                metadata_data = metadata_raw.get('data')            # todo: use native Memory_FS methods here (and we should be using a Type_Safe class here)

            # Get content encoding from metadata
            content_encoding = metadata_data.get('content_encoding')

            # Handle decompression if needed
            if content_encoding == 'gzip' and isinstance(data, bytes):
                data = gzip.decompress(data)
                # After decompression, determine if it's JSON or remains binary
                try:                                            # todo: we should know this from the metadata/config
                    data = json.loads(data.decode('utf-8'))
                    data_type = "json"
                except (json.JSONDecodeError, UnicodeDecodeError):
                    data_type = "binary"
            else:
                data_type = self.determine_data_type(data)

            return { "data"            : data             ,                     # ths should be a Type_Safe class
                     "metadata"        : metadata_data    ,
                     "data_type"       : data_type        ,
                     "content_encoding": content_encoding }

        return None

    def retrieve_by_id__config(self, cache_id  : Cache_Id,
                                     namespace : Safe_Str__Id
                                ) -> Schema__Memory_FS__File__Config:
        file_refs = self.retrieve_by_id__refs(cache_id, namespace)
        if not file_refs or not file_refs.file_paths.content_files:
            return None

        handler      = self.get_or_create_handler(namespace)
        for content_path in file_refs.file_paths.content_files:
            config_path  = content_path + '.config'
            config_json  = handler.storage_backend.file__json(config_path)

            if not config_json:
                return None

            return Schema__Memory_FS__File__Config.from_json(config_json)
        return None

    # todo: this (and similar methods) need to be refactored so that we don't  mix the concerns
    #       at the moment this method is both finding the find and doing the cast to Schema__Cache__File__Metadata
    def retrieve_by_id__metadata(self, cache_id  : Cache_Id,
                                   namespace : Safe_Str__Id
                              ) -> Schema__Cache__File__Metadata:
        file_refs = self.retrieve_by_id__refs(cache_id, namespace)                  # Step 1: Use existing helper to get refs
        if not file_refs or not file_refs.file_paths.content_files:
            return None

        handler       = self.get_or_create_handler(namespace)                       # Step 2: Read metadata from the DATA file
        for content_path in file_refs.file_paths.content_files:
            metadata_path = content_path + '.metadata'
            metadata_json = handler.storage_backend.file__json(metadata_path)

            if not metadata_json:
                return None

            return Schema__Cache__File__Metadata.from_json(metadata_json)
        return None

    def retrieve_by_id__refs(self, cache_id  : Cache_Id,
                                   namespace : Safe_Str__Id
                                ) -> Schema__Cache__File__Refs:                      #   Retrieve by cache ID using direct path from reference
        if cache_id:
            handler   = self.get_or_create_handler(namespace)
            with handler.fs__refs_id.file__json__single(Safe_Str__Id(cache_id)) as ref_fs:           # get the main by-id file, which contains pointers to the other files
                json_data = ref_fs.content()                                                 # todo refactor this so that we get the Schema__Cache__File__Refs directly from fs__refs_id
                if json_data:
                    return Schema__Cache__File__Refs.from_json(json_data)
        return None

    def determine_data_type(self, data) -> str:
        if isinstance(data, bytes):
            return "binary"
        elif isinstance(data, dict) or isinstance(data, list):
            return "json"
        else:
            return "string"

    @type_safe
    def hash_from_string(self, data: str) -> Safe_Str__Cache__File__Cache_Hash:                       # Calculate hash from string
        return self.hash_generator.from_string(data)

    @type_safe
    def hash_from_bytes(self, data: bytes) -> Safe_Str__Cache__File__Cache_Hash:                      # Calculate hash from bytes
        return self.hash_generator.from_bytes(data)

    @type_safe
    def hash_from_json(self, data          : dict        ,                                            # Calculate hash from JSON
                             exclude_fields : List[str] = None
                        ) -> Safe_Str__Cache__File__Cache_Hash:
        return self.hash_generator.from_json(data, exclude_fields)

    def hash_from_json_field(self, data      : dict,
                                   json_field: Safe_Str__Json__Field_Path,
                              ) -> Safe_Str__Cache__File__Cache_Hash:
        return self.hash_generator.from_json_field(data, json_field)


from typing                                                                                      import Any, Optional
from osbot_utils.decorators.methods.cache_on_self                                                import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                             import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                               import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                  import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                   import type_safe
from osbot_utils.utils.Http                                                                      import url_join_safe
from osbot_utils.utils.Json                                                                      import json_to_bytes
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Update__Request      import Schema__Cache__Data__Update__Request
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Update__Response     import Schema__Cache__Data__Update__Response
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                   import Enum__Cache__Data_Type
from mgraph_ai_service_cache.service.cache.Cache__Service                                        import Cache__Service
from mgraph_ai_service_cache.service.cache.retrieve.Cache__Service__Retrieve                     import Cache__Service__Retrieve


class Cache__Service__Data__Update(Type_Safe):                                                  # Service for updating data files in cache entries
    cache_service : Cache__Service

    @cache_on_self
    def retrieve_service(self):
        return Cache__Service__Retrieve(cache_service=self.cache_service)

    def get_extension_for_type(self, data_type: Enum__Cache__Data_Type) -> str:                 # Map data type to file extension
        extensions = { Enum__Cache__Data_Type.STRING : 'txt' ,
                       Enum__Cache__Data_Type.JSON   : 'json',
                       Enum__Cache__Data_Type.BINARY : 'bin' }
        return extensions.get(data_type, 'bin')

    @type_safe
    def build_data_file_path(self, data_folder  : Safe_Str__File__Path,                         # Centralized path construction matching other data services
                                   data_key     : Safe_Str__File__Path = None,
                                   data_file_id : Safe_Str__Id         = None,
                                   data_type    : Enum__Cache__Data_Type = None
                            ) -> Safe_Str__File__Path:

        if not data_file_id:
            raise ValueError("data_file_id is required for data file operations")

        path = data_folder                                                                      # Start with base data folder

        if data_key:                                                                            # Add optional data_key path segment
            path = url_join_safe(path, data_key)

        path = url_join_safe(path, data_file_id)                                                # Add data_file_id

        if data_type:                                                                           # Add extension based on data type
            extension = self.get_extension_for_type(data_type)
            path = Safe_Str__File__Path(f"{path}.{extension}")

        return path

    def serialize_data(self, data: Any, data_type: Enum__Cache__Data_Type) -> bytes:            # Serialize data based on its type
        if data_type == Enum__Cache__Data_Type.STRING:
            return str(data).encode('utf-8')
        elif data_type == Enum__Cache__Data_Type.JSON:
            if isinstance(data, bytes):
                return data
            return json_to_bytes(data)
        elif data_type == Enum__Cache__Data_Type.BINARY:
            if isinstance(data, bytes):
                return data
            raise ValueError(f"Binary data must be bytes, got {type(data)}")
        else:
            raise ValueError(f"Unknown data type: {data_type}")

    @type_safe
    def update_data(self, request: Schema__Cache__Data__Update__Request
                   ) -> Optional[Schema__Cache__Data__Update__Response]:                        # Update an existing data file under a cache entry

        file_refs = self.retrieve_service().retrieve_by_id__refs(cache_id  = request.cache_id,
                                                                 namespace = request.namespace)

        if not file_refs or not file_refs.file_paths.data_folders:
            return None                                                                         # Cache entry doesn't exist

        handler       = self.cache_service.get_or_create_handler(request.namespace)
        updated_paths = []

        for data_folder in file_refs.file_paths.data_folders:                                   # Try each data folder
            full_path = self.build_data_file_path(data_folder   = data_folder          ,
                                                  data_key      = request.data_key     ,
                                                  data_file_id  = request.data_file_id ,
                                                  data_type     = request.data_type    )

            if handler.storage_backend.file__exists(full_path):                                 # Only update if file exists
                serialized_data = self.serialize_data(request.data, request.data_type)
                success = handler.storage_backend.file__save(full_path, serialized_data)

                if success:
                    updated_paths.append(str(full_path))

        if not updated_paths:                                                                   # File not found in any data folder
            return None

        serialized_data = self.serialize_data(request.data, request.data_type)                  # Calculate file size
        file_size       = len(serialized_data)

        return Schema__Cache__Data__Update__Response(success      = True                ,
                                                     cache_id     = request.cache_id    ,
                                                     namespace    = request.namespace   ,
                                                     data_type    = request.data_type   ,
                                                     data_key     = request.data_key    ,
                                                     data_file_id = request.data_file_id,
                                                     file_path    = updated_paths[0]    ,       # Return first updated path
                                                     file_size    = file_size           )


from typing import Type, TypeVar, Optional
from enum import Enum
from osbot_utils.type_safe.Type_Safe__Primitive import Type_Safe__Primitive
from osbot_utils.utils.Env import get_env

T = TypeVar('T', bound=Enum)
P = TypeVar('P', bound=Type_Safe__Primitive)

def get_env_enum(env_var_name : str                ,                            # Environment variable name to look up
                 enum_class   : Type[T]            ,                            # Enum class to convert to
                 default      : Optional[T] = None                              # Optional default if env var not set or invalid
            ) -> Optional[T]:                                                   # Returns enum value or default
    env_value = get_env(env_var_name)
    if not env_value:
        return default

    try:                                                                        # Convert string to appropriate type based on enum's base class
        if issubclass(enum_class, int):                                         # Integer-based enums
            env_value = int(env_value)
        elif issubclass(enum_class, float) and not issubclass(enum_class, int): # Float-based enums (for numeric thresholds, etc.)
            env_value = float(env_value)
        elif issubclass(enum_class, bytes):                                     # Bytes-based enums (rare, for binary protocols)
            env_value = env_value.encode('utf-8')
        else:
            pass                                                                # str enums pass through as-is (most common case)

        return enum_class(env_value)
    except (ValueError, KeyError, UnicodeEncodeError):
        return default

def get_env_primitive(env_var_name  : str                     ,                   # Environment variable name to look up
                      primitive_class: Type[P]                ,                   # Type_Safe__Primitive class to convert to
                      default        : Optional[P] = None                         # Optional default if env var not set or invalid
                      ) -> Optional[P]:                                           # Returns primitive value or default
    env_value = get_env(env_var_name)
    if not env_value:
        return default

    try:
        return primitive_class(env_value)
    except (ValueError, TypeError):
        return default


from typing                                                                                 import Dict, Union
from fastapi                                                                                import HTTPException, Body
from osbot_fast_api.api.decorators.route_path                                               import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                             import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                  import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                     import Safe_Str__Fast_API__Route__Tag
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__File__Data_Key  import Safe_Str__Cache__File__Data_Key
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__File__File_Id   import Safe_Str__Cache__File__File_Id
from mgraph_ai_service_cache_client.schemas.cache.safe_str.Safe_Str__Cache__Namespace       import Safe_Str__Cache__Namespace
from osbot_utils.decorators.methods.cache_on_self                                           import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                        import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path           import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                          import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id             import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                              import type_safe
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Store__Request  import Schema__Cache__Data__Store__Request
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Store__Response import Schema__Cache__Data__Store__Response
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type              import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                          import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                   import Cache__Service
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__Store                 import Cache__Service__Data__Store


TAG__ROUTES_STORE__DATA       = Safe_Str__Fast_API__Route__Tag('data/store')
PREFIX__ROUTES_STORE__DATA    = Safe_Str__Fast_API__Route__Prefix('/{namespace}/cache/{cache_id}')
BASE_PATH__ROUTES_STORE__DATA = f'{PREFIX__ROUTES_STORE__DATA}/{TAG__ROUTES_STORE__DATA}/'

ROUTES_PATHS__STORE__DATA = [ BASE_PATH__ROUTES_STORE__DATA + 'string'                               ,
                              BASE_PATH__ROUTES_STORE__DATA + 'string/{data_file_id}'                ,
                              BASE_PATH__ROUTES_STORE__DATA + 'string/{data_key:path}/{data_file_id}',
                              BASE_PATH__ROUTES_STORE__DATA + 'json'                                 ,
                              BASE_PATH__ROUTES_STORE__DATA + 'json/{data_file_id}'                  ,
                              BASE_PATH__ROUTES_STORE__DATA + 'json/{data_key:path}/{data_file_id}'  ,
                              BASE_PATH__ROUTES_STORE__DATA + 'binary'                               ,
                              BASE_PATH__ROUTES_STORE__DATA + 'binary/{data_file_id}'                ,
                              BASE_PATH__ROUTES_STORE__DATA + 'binary/{data_key:path}/{data_file_id}']


class Routes__Data__Store(Fast_API__Routes):                                            # FastAPI routes for storing data files as cache entry children
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_STORE__DATA
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_STORE__DATA
    cache_service : Cache__Service                                                      # Dependency injection for cache service

    @cache_on_self
    def store_service(self) -> Cache__Service__Data__Store:                             # Service layer for data storage logic
        return Cache__Service__Data__Store(cache_service=self.cache_service)

    @route_path("/data/store/binary")
    def data__store_binary(self, body: bytes         = Body(...)                  ,
                                 cache_id     : Cache_Id      = None                       ,
                                 namespace    : Safe_Str__Id  = FAST_API__PARAM__NAMESPACE ,
                            ) -> Schema__Cache__Data__Store__Response:
        return self.data__store_binary__with__id_and_key(body         = body        ,
                                                         cache_id     = cache_id    ,
                                                         namespace    = namespace   ,
                                                         data_key     = ''          ,
                                                         data_file_id = ''          )

    @route_path("/data/store/binary/{data_file_id}")
    def data__store_binary__with__id(self, body         : bytes         = Body(...)                     ,
                                           cache_id     : Cache_Id      = None                          ,
                                           namespace    : Safe_Str__Id  = FAST_API__PARAM__NAMESPACE    ,
                                           data_file_id : Safe_Str__Id  = None
                                    ) -> Schema__Cache__Data__Store__Response:
        return self.data__store_binary__with__id_and_key(body         = body        ,
                                                         cache_id     = cache_id    ,
                                                         namespace    = namespace   ,
                                                         data_key     = ''          ,
                                                         data_file_id = data_file_id)

    @route_path("/data/store/binary/{data_key:path}/{data_file_id}")
    def data__store_binary__with__id_and_key(self, body         : bytes                  = Body(..., media_type="application/octet-stream"),
                                                   cache_id     : Cache_Id               = None                          ,
                                                   namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                                   data_key     : Safe_Str__File__Path   = None                          ,
                                                   data_file_id : Safe_Str__Id           = None
                                           ) -> Schema__Cache__Data__Store__Response:                                    # Store binary data file under cache entry

        request = Schema__Cache__Data__Store__Request(cache_id     = cache_id                     ,
                                                      data         = body                         ,
                                                      data_type    = Enum__Cache__Data_Type.BINARY,
                                                      data_key     = data_key                     ,
                                                      data_file_id = data_file_id                 ,
                                                      namespace    = namespace                    )

        result = self.store_service().store_data(request)
        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)

    @route_path("/data/store/json")
    def data__store_json(self, data         : dict         = Body(...)                  ,
                               cache_id     : Cache_Id     = None                       ,
                               namespace    : Safe_Str__Id = FAST_API__PARAM__NAMESPACE ,
                        ) -> Schema__Cache__Data__Store__Response:
        return self.data__store_json__with__id_and_key(data         = data      ,
                                                       cache_id     = cache_id  ,
                                                       namespace    = namespace ,
                                                       data_key     = ''        ,
                                                       data_file_id = ''        )

    @route_path("/data/store/json/{data_file_id}")
    def data__store_json__with__id(self, data         : dict         = Body(...)                  ,
                                         cache_id     : Cache_Id     = None                       ,
                                         namespace    : Safe_Str__Id = FAST_API__PARAM__NAMESPACE ,
                                         data_file_id : Safe_Str__Id = None
                                  ) -> Schema__Cache__Data__Store__Response:
        return self.data__store_json__with__id_and_key(data           = data        ,
                                                      cache_id     = cache_id      ,
                                                      namespace    = namespace     ,
                                                      data_key     = ''            ,
                                                      data_file_id = data_file_id  )

    @route_path("/data/store/json/{data_key:path}/{data_file_id}")
    def data__store_json__with__id_and_key(self, data         : dict                            = Body(...)                     ,
                                                 cache_id     : Cache_Id                        = None                          ,
                                                 namespace    : Safe_Str__Cache__Namespace      = FAST_API__PARAM__NAMESPACE    ,
                                                 data_key     : Safe_Str__Cache__File__Data_Key = None                          ,
                                                 data_file_id : Safe_Str__Cache__File__File_Id  = None
                                         ) -> Schema__Cache__Data__Store__Response:                                     # Store JSON data file under cache entry
        request = Schema__Cache__Data__Store__Request(cache_id     = cache_id              ,
                                                      data         = data                  ,
                                                      data_type    = Enum__Cache__Data_Type.JSON,
                                                      data_key     = data_key              ,
                                                      data_file_id = data_file_id          ,
                                                      namespace    = namespace             )

        result = self.store_service().store_data(request)

        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)

    def test_404(self, cache_id     : Cache_Id               = None                          ,
                       namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,):

        error_detail = { "error_type" : "NOT_FOUND"                                                      ,
                             "message"    : f"Cache entry '{cache_id}' in namespace '{namespace}' not found" ,
                             "cache_id"   : str(cache_id)                                                    }
        raise HTTPException(status_code=404, detail=error_detail)

    @route_path("/data/store/string")
    def data__store_string(self, data         : str                    = Body(...)                     ,
                                  cache_id     : Cache_Id               = None                          ,
                                  namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                          ) -> Schema__Cache__Data__Store__Response:

        return self.data__store_string__with__id_and_key(data         = data        ,
                                                         cache_id     = cache_id    ,
                                                         namespace    = namespace   ,
                                                         data_key     = ''          ,
                                                         data_file_id = ''          )

    @route_path("/data/store/string/{data_file_id}")
    def data__store_string__with__id(self, data         : str                    = Body(...)                     ,
                                           cache_id     : Cache_Id               = None                          ,
                                           namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                           data_file_id : Safe_Str__Id           = None
                                    ) -> Schema__Cache__Data__Store__Response:
        return self.data__store_string__with__id_and_key(data         = data        ,
                                                         cache_id     = cache_id    ,
                                                         namespace    = namespace   ,
                                                         data_key     = ''          ,
                                                         data_file_id = data_file_id)

    @route_path("/data/store/string/{data_key:path}/{data_file_id}")
    def data__store_string__with__id_and_key(self, data         : str                    = Body(...)                     ,
                                           cache_id     : Cache_Id               = None                          ,
                                           namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                           data_key     : Safe_Str__File__Path   = None                          ,
                                           data_file_id : Safe_Str__Id           = None
                                    ) -> Schema__Cache__Data__Store__Response:                                   # Store string data file under cache entry

        request = Schema__Cache__Data__Store__Request(cache_id     = cache_id                ,
                                                      data         = data                    ,
                                                      data_type    = Enum__Cache__Data_Type.STRING,
                                                      data_key     = data_key                ,
                                                      data_file_id = data_file_id            ,
                                                      namespace    = namespace               )
        result = self.store_service().store_data(request)

        return self.handle_not_found(result, cache_id=cache_id, namespace=namespace)

    @type_safe
    def handle_not_found(self, result        : Union[Type_Safe, Dict] = None,                                 # Base method for 404 handling
                               cache_id      : Cache_Id               = None,
                               namespace     : Safe_Str__Id           = None):
        if result is None:
            error_detail = { "error_type" : "NOT_FOUND"                                                      ,
                             "message"    : f"Cache entry '{cache_id}' in namespace '{namespace}' not found" ,
                             "cache_id"   : str(cache_id)                                                    }
            raise HTTPException(status_code=404, detail=error_detail)
        return result

    def setup_routes(self):                                                             # Configure all data storage routes
        self.add_route_post(self.data__store_binary                   )                 # binary endpoints
        self.add_route_post(self.data__store_binary__with__id         )
        self.add_route_post(self.data__store_binary__with__id_and_key )

        self.add_route_post(self.data__store_json                     )                 # json endpoints
        self.add_route_post(self.data__store_json__with__id           )
        self.add_route_post(self.data__store_json__with__id_and_key   )

        self.add_route_post(self.data__store_string                   )                 # string endpoints
        self.add_route_post(self.data__store_string__with__id         )
        self.add_route_post(self.data__store_string__with__id_and_key )


from fastapi                                                                                      import HTTPException
from osbot_fast_api.api.decorators.route_path                                                     import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                   import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                        import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                           import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                                 import cache_on_self
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                 import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                                import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                   import Safe_Str__Id
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                    import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                                import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                         import Cache__Service
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__Exists                      import Cache__Service__Data__Exists
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Exists__Response      import Schema__Cache__Data__Exists__Response

TAG__ROUTES_EXISTS__DATA       = Safe_Str__Fast_API__Route__Tag('data')
PREFIX__ROUTES_EXISTS__DATA    = Safe_Str__Fast_API__Route__Prefix('/{namespace}/cache/{cache_id}')
BASE_PATH__ROUTES_EXISTS__DATA = f'{PREFIX__ROUTES_EXISTS__DATA}/{TAG__ROUTES_EXISTS__DATA}/'

ROUTES_PATHS__EXISTS__DATA = [ BASE_PATH__ROUTES_EXISTS__DATA + 'exists/{data_type}/{data_file_id}'                  ,
                               BASE_PATH__ROUTES_EXISTS__DATA + 'exists/{data_type}/{data_key:path}/{data_file_id}' ]


class Routes__Data__Exists(Fast_API__Routes):                                                   # FastAPI routes for checking if data files exist in cache entries
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_EXISTS__DATA
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_EXISTS__DATA
    cache_service : Cache__Service                                                              # Dependency injection for cache service

    @cache_on_self
    def exists_service(self) -> Cache__Service__Data__Exists:                                   # Service layer for exists operations
        return Cache__Service__Data__Exists(cache_service=self.cache_service)

    @route_path("/data/exists/{data_type}/{data_file_id}")
    def data__exists__with__id(self, cache_id     : Cache_Id               = None                          ,
                                     namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                     data_type    : Enum__Cache__Data_Type = None                          ,
                                     data_file_id : Safe_Str__Id           = None
                               ) -> Schema__Cache__Data__Exists__Response:
        return self.data__exists__with__id_and_key(cache_id     = cache_id    ,
                                                   namespace    = namespace   ,
                                                   data_type    = data_type   ,
                                                   data_key     = ''          ,
                                                   data_file_id = data_file_id)

    @route_path("/data/exists/{data_type}/{data_key:path}/{data_file_id}")
    def data__exists__with__id_and_key(self, cache_id     : Cache_Id               = None                          ,
                                             namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                             data_type    : Enum__Cache__Data_Type = None                          ,
                                             data_key     : Safe_Str__File__Path   = None                          ,
                                             data_file_id : Safe_Str__Id           = None
                                       ) -> Schema__Cache__Data__Exists__Response:                                  # Check if specific data file exists

        if not data_file_id:
            error_detail = { "error_type" : "INVALID_INPUT",
                            "message"    : "data_file_id is required to check existence"}
            raise HTTPException(status_code=400, detail=error_detail)

        if not data_type:
            error_detail = { "error_type" : "INVALID_INPUT",
                            "message"    : "data_type is required to check existence"}
            raise HTTPException(status_code=400, detail=error_detail)

        exists = self.exists_service().data_file_exists(cache_id     = cache_id    ,
                                                        namespace    = namespace   ,
                                                        data_type    = data_type   ,
                                                        data_key     = data_key    ,
                                                        data_file_id = data_file_id)

        return Schema__Cache__Data__Exists__Response(exists       = exists      ,
                                                     cache_id     = cache_id    ,
                                                     namespace    = namespace   ,
                                                     data_type    = data_type   ,
                                                     data_key     = data_key    ,
                                                     data_file_id = data_file_id)

    def setup_routes(self):                                                                     # Configure all data exists routes
        self.add_route_get(self.data__exists__with__id        )
        self.add_route_get(self.data__exists__with__id_and_key)


from typing                                                                                       import List
from fastapi                                                                                      import HTTPException
from osbot_fast_api.api.decorators.route_path                                                     import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                   import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                        import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                           import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                                 import cache_on_self
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                 import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                                import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                   import Safe_Str__Id
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                                import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                         import Cache__Service
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__List                        import Cache__Service__Data__List
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__List__Response        import Schema__Cache__Data__List__Response

TAG__ROUTES_LIST__DATA       = Safe_Str__Fast_API__Route__Tag('data')
PREFIX__ROUTES_LIST__DATA    = Safe_Str__Fast_API__Route__Prefix('/{namespace}/cache/{cache_id}')
BASE_PATH__ROUTES_LIST__DATA = f'{PREFIX__ROUTES_LIST__DATA}/{TAG__ROUTES_LIST__DATA}/'

ROUTES_PATHS__LIST__DATA = [ BASE_PATH__ROUTES_LIST__DATA + 'list'                ,
                             BASE_PATH__ROUTES_LIST__DATA + 'list/{data_key:path}']


class Routes__Data__List(Fast_API__Routes):                                                     # FastAPI routes for listing data files in cache entries
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_LIST__DATA
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_LIST__DATA
    cache_service : Cache__Service                                                              # Dependency injection for cache service

    @cache_on_self
    def list_service(self) -> Cache__Service__Data__List:                                       # Service layer for list operations
        return Cache__Service__Data__List(cache_service=self.cache_service)

    @route_path("/data/list")
    def data__list(self, cache_id  : Cache_Id     = None                          ,
                         namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE    ,
                         recursive : bool         = True
                   ) -> Schema__Cache__Data__List__Response:
        return self.data__list__with__key(cache_id  = cache_id  ,
                                          namespace = namespace ,
                                          data_key  = ''        ,
                                          recursive = recursive )

    @route_path("/data/list/{data_key:path}")
    def data__list__with__key(self, cache_id  : Cache_Id             = None                          ,
                                    namespace : Safe_Str__Id         = FAST_API__PARAM__NAMESPACE    ,
                                    data_key  : Safe_Str__File__Path = None                          ,
                                    recursive : bool                 = True
                              ) -> Schema__Cache__Data__List__Response:                         # List all data files under cache entry

        result = self.list_service().list_data_files(cache_id  = cache_id  ,
                                                     namespace = namespace ,
                                                     data_key  = data_key  ,
                                                     recursive = recursive )

        if result is None:
            error_detail = { "error_type" : "NOT_FOUND"                                        ,
                             "message"    : f"Cache entry '{cache_id}' not found"              ,
                             "cache_id"   : str(cache_id)                                      }
            raise HTTPException(status_code=404, detail=error_detail)

        return result

    def setup_routes(self):                                                                     # Configure all data list routes
        self.add_route_get(self.data__list         )
        self.add_route_get(self.data__list__with__key)


from osbot_fast_api.utils.Fast_API__Server_Info                            import fast_api__server_info, Fast_API__Server_Info
from osbot_utils.type_safe.Type_Safe                                       import Type_Safe
from mgraph_ai_service_cache.service.info.schemas.Schema__Service__Status   import Schema__Service__Status, Enum__Service_Environment
from mgraph_ai_service_cache.service.info.schemas.Schema__Server__Versions  import Schema__Server__Versions


class Service_Info(Type_Safe):

    def environment(self):                                                   # Determine current environment
        import os
        if os.getenv('AWS_REGION'):
            return Enum__Service_Environment.aws_lambda
        else:
            return Enum__Service_Environment.local

    def service_info(self) -> Schema__Service__Status:                               # Get current service status
        return Schema__Service__Status(environment = self.environment())

    def versions(self):
        return Schema__Server__Versions()

    def server_info(self) -> Fast_API__Server_Info:
        return fast_api__server_info



from osbot_utils.decorators.methods.cache_on_self                                   import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path   import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                  import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id     import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                      import type_safe
from osbot_utils.utils.Http                                                         import url_join_safe
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type      import Enum__Cache__Data_Type
from mgraph_ai_service_cache.service.cache.Cache__Service                           import Cache__Service
from mgraph_ai_service_cache.service.cache.retrieve.Cache__Service__Retrieve        import Cache__Service__Retrieve


class Cache__Service__Data__Exists(Type_Safe):                                              # Service for checking if data files exist in cache entries
    cache_service : Cache__Service

    @cache_on_self
    def retrieve_service(self):
        return Cache__Service__Retrieve(cache_service=self.cache_service)

    def get_extension_for_type(self, data_type: Enum__Cache__Data_Type) -> str:             # Map data type to file extension
        extensions = { Enum__Cache__Data_Type.STRING : 'txt' ,
                       Enum__Cache__Data_Type.JSON   : 'json',
                       Enum__Cache__Data_Type.BINARY : 'bin' }
        return extensions.get(data_type, 'bin')

    @type_safe
    def build_data_file_path(self, data_folder  : Safe_Str__File__Path,                     # Centralized path construction matching other data services
                                   data_key     : Safe_Str__File__Path = None,
                                   data_file_id : Safe_Str__Id         = None,
                                   data_type    : Enum__Cache__Data_Type = None
                            ) -> Safe_Str__File__Path:

        if not data_file_id:
            raise ValueError("data_file_id is required for data file operations")

        path = data_folder                                                                  # Start with base data folder

        if data_key:                                                                        # Add optional data_key path segment
            path = url_join_safe(path, data_key)

        path = url_join_safe(path, data_file_id)                                            # Add data_file_id

        if data_type:                                                                       # Add extension based on data type
            extension = self.get_extension_for_type(data_type)
            path = Safe_Str__File__Path(f"{path}.{extension}")

        return path

    @type_safe
    def data_file_exists(self,
                         cache_id     : Cache_Id              ,
                         namespace    : Safe_Str__Id          ,
                         data_type    : Enum__Cache__Data_Type,
                         data_key     : Safe_Str__File__Path  ,
                         data_file_id : Safe_Str__Id
                    ) -> bool:                                                              # Check if a specific data file exists under a cache entry

        file_refs = self.retrieve_service().retrieve_by_id__refs(cache_id  = cache_id,
                                                                 namespace = namespace)

        if not file_refs or not file_refs.file_paths.data_folders:
            return False

        handler = self.cache_service.get_or_create_handler(namespace)

        for data_folder in file_refs.file_paths.data_folders:                               # Check each data folder
            full_path = self.build_data_file_path(data_folder   = data_folder   ,
                                                  data_key      = data_key      ,
                                                  data_file_id  = data_file_id  ,
                                                  data_type     = data_type     )

            if handler.storage_backend.file__exists(full_path):                             # file__exists is available in Storage_FS
                return True

        return False


from typing                                                                                       import Union, Dict
from fastapi                                                                                      import HTTPException, Response, Body
from osbot_fast_api.api.decorators.route_path                                                     import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                   import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                        import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                           import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                                 import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                 import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                                import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                   import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                                    import type_safe
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Update__Request       import Schema__Cache__Data__Update__Request
from mgraph_ai_service_cache_client.schemas.cache.data.Schema__Cache__Data__Update__Response      import Schema__Cache__Data__Update__Response
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Data_Type                    import Enum__Cache__Data_Type
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                                import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                         import Cache__Service
from mgraph_ai_service_cache.service.cache.data.Cache__Service__Data__Update                      import Cache__Service__Data__Update

TAG__ROUTES_UPDATE__DATA       = Safe_Str__Fast_API__Route__Tag('data/update')
PREFIX__ROUTES_UPDATE__DATA    = Safe_Str__Fast_API__Route__Prefix('/{namespace}/cache/{cache_id}')
BASE_PATH__ROUTES_UPDATE__DATA = f'{PREFIX__ROUTES_UPDATE__DATA}/{TAG__ROUTES_UPDATE__DATA}/'

ROUTES_PATHS__UPDATE__DATA = [ BASE_PATH__ROUTES_UPDATE__DATA + 'string/{data_file_id}'                ,
                               BASE_PATH__ROUTES_UPDATE__DATA + 'string/{data_key:path}/{data_file_id}',
                               BASE_PATH__ROUTES_UPDATE__DATA + 'json/{data_file_id}'                  ,
                               BASE_PATH__ROUTES_UPDATE__DATA + 'json/{data_key:path}/{data_file_id}'  ,
                               BASE_PATH__ROUTES_UPDATE__DATA + 'binary/{data_file_id}'                ,
                               BASE_PATH__ROUTES_UPDATE__DATA + 'binary/{data_key:path}/{data_file_id}']


class Routes__Data__Update(Fast_API__Routes):                                                   # FastAPI routes for updating data files in cache entries
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_UPDATE__DATA
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_UPDATE__DATA
    cache_service : Cache__Service                                                              # Dependency injection for cache service

    @cache_on_self
    def update_service(self) -> Cache__Service__Data__Update:                                   # Service layer for update operations
        return Cache__Service__Data__Update(cache_service=self.cache_service)

    # String update endpoints

    @route_path("/data/update/string/{data_file_id}")
    def data__update_string__with__id(self, data         : str          = Body(...)                     ,
                                            cache_id     : Cache_Id     = None                          ,
                                            namespace    : Safe_Str__Id = FAST_API__PARAM__NAMESPACE    ,
                                            data_file_id : Safe_Str__Id = None
                                      ) -> Schema__Cache__Data__Update__Response:
        return self.data__update_string__with__id_and_key(data         = data        ,
                                                          cache_id     = cache_id    ,
                                                          namespace    = namespace   ,
                                                          data_key     = ''          ,
                                                          data_file_id = data_file_id)

    @route_path("/data/update/string/{data_key:path}/{data_file_id}")
    def data__update_string__with__id_and_key(self, data         : str                    = Body(...)                     ,
                                                    cache_id     : Cache_Id               = None                          ,
                                                    namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                                    data_key     : Safe_Str__File__Path   = None                          ,
                                                    data_file_id : Safe_Str__Id           = None
                                              ) -> Schema__Cache__Data__Update__Response:       # Update string data file

        if not data:
            error_detail = { "error_type" : "INVALID_INPUT"             ,
                            "message"    : "String data cannot be empty",
                            "field_name" : "data"                       }
            raise HTTPException(status_code=400, detail=error_detail)

        request = Schema__Cache__Data__Update__Request(cache_id     = cache_id                ,
                                                       data         = data                    ,
                                                       data_type    = Enum__Cache__Data_Type.STRING,
                                                       data_key     = data_key                ,
                                                       data_file_id = data_file_id            ,
                                                       namespace    = namespace               )

        result = self.update_service().update_data(request)
        return self.handle_not_found(result, cache_id=cache_id, data_file_id=data_file_id)

    # JSON update endpoints

    @route_path("/data/update/json/{data_file_id}")
    def data__update_json__with__id(self, data         : dict         = Body(...)                     ,
                                          cache_id     : Cache_Id     = None                          ,
                                          namespace    : Safe_Str__Id = FAST_API__PARAM__NAMESPACE    ,
                                          data_file_id : Safe_Str__Id = None
                                    ) -> Schema__Cache__Data__Update__Response:
        return self.data__update_json__with__id_and_key(data         = data        ,
                                                        cache_id     = cache_id    ,
                                                        namespace    = namespace   ,
                                                        data_key     = ''          ,
                                                        data_file_id = data_file_id)

    @route_path("/data/update/json/{data_key:path}/{data_file_id}")
    def data__update_json__with__id_and_key(self, data         : dict                   = Body(...)                     ,
                                                  cache_id     : Cache_Id               = None                          ,
                                                  namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                                  data_key     : Safe_Str__File__Path   = None                          ,
                                                  data_file_id : Safe_Str__Id           = None
                                            ) -> Schema__Cache__Data__Update__Response:         # Update JSON data file

        request = Schema__Cache__Data__Update__Request(cache_id     = cache_id              ,
                                                       data         = data                  ,
                                                       data_type    = Enum__Cache__Data_Type.JSON,
                                                       data_key     = data_key              ,
                                                       data_file_id = data_file_id          ,
                                                       namespace    = namespace             )

        result = self.update_service().update_data(request)
        return self.handle_not_found(result, cache_id=cache_id, data_file_id=data_file_id)

    # Binary update endpoints

    @route_path("/data/update/binary/{data_file_id}")
    def data__update_binary__with__id(self, body         : bytes        = Body(..., media_type="application/octet-stream"),
                                            cache_id     : Cache_Id     = None                          ,
                                            namespace    : Safe_Str__Id = FAST_API__PARAM__NAMESPACE    ,
                                            data_file_id : Safe_Str__Id = None
                                      ) -> Schema__Cache__Data__Update__Response:
        return self.data__update_binary__with__id_and_key(body         = body        ,
                                                          cache_id     = cache_id    ,
                                                          namespace    = namespace   ,
                                                          data_key     = ''          ,
                                                          data_file_id = data_file_id)

    @route_path("/data/update/binary/{data_key:path}/{data_file_id}")
    def data__update_binary__with__id_and_key(self, body         : bytes                  = Body(..., media_type="application/octet-stream"),
                                                    cache_id     : Cache_Id               = None                          ,
                                                    namespace    : Safe_Str__Id           = FAST_API__PARAM__NAMESPACE    ,
                                                    data_key     : Safe_Str__File__Path   = None                          ,
                                                    data_file_id : Safe_Str__Id           = None
                                              ) -> Schema__Cache__Data__Update__Response:       # Update binary data file

        if not body:
            error_detail = { "error_type" : "INVALID_INPUT"             ,
                            "message"    : "Binary data cannot be empty",
                            "field_name" : "body"                       }
            raise HTTPException(status_code=400, detail=error_detail)

        request = Schema__Cache__Data__Update__Request(cache_id     = cache_id                 ,
                                                       data         = body                     ,
                                                       data_type    = Enum__Cache__Data_Type.BINARY,
                                                       data_key     = data_key                 ,
                                                       data_file_id = data_file_id             ,
                                                       namespace    = namespace                )

        result = self.update_service().update_data(request)
        return self.handle_not_found(result, cache_id=cache_id, data_file_id=data_file_id)

    @type_safe
    def handle_not_found(self, result        : Union[Type_Safe, Dict] = None,                   # Base method for 404 handling
                               cache_id      : Cache_Id               = None,
                               data_file_id  : Safe_Str__Id           = None):
        if result is None:
            error_detail = { "error_type"   : "NOT_FOUND"                                      ,
                             "message"      : f"Data file '{data_file_id}' not found in cache entry '{cache_id}'",
                             "cache_id"     : str(cache_id)                                    ,
                             "data_file_id" : str(data_file_id)                                }
            raise HTTPException(status_code=404, detail=error_detail)
        return result

    def setup_routes(self):                                                                     # Configure all data update routes
        self.add_route_post(self.data__update_string__with__id        )                         # string endpoints
        self.add_route_post(self.data__update_string__with__id_and_key)

        self.add_route_post(self.data__update_json__with__id          )                         # json endpoints
        self.add_route_post(self.data__update_json__with__id_and_key  )

        self.add_route_post(self.data__update_binary__with__id        )                         # binary endpoints
        self.add_route_post(self.data__update_binary__with__id_and_key)


from osbot_utils.type_safe.Type_Safe                                                      import Type_Safe
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                            import type_safe
from osbot_utils.utils.Zip                                                                import zip_bytes__file_list
from mgraph_ai_service_cache.service.cache.Cache__Service                                 import Cache__Service
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Store__Request  import Schema__Cache__Zip__Store__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Store__Response import Schema__Cache__Zip__Store__Response
from mgraph_ai_service_cache.utils.for_osbot_utils.Zip                                    import  zip_bytes__content_hash


class Cache__Service__Zip__Store(Type_Safe):                                            # Service layer for storing zip files
    cache_service : Cache__Service                                                      # Underlying cache service

    @type_safe
    def store_zip(self, request: Schema__Cache__Zip__Store__Request
                   ) -> Schema__Cache__Zip__Store__Response:                             # Store zip file in cache

        if not request.zip_bytes:
            return Schema__Cache__Zip__Store__Response(success       = False,
                                                       namespace     = request.namespace,
                                                       error_type    = "INVALID_INPUT",
                                                       error_message = "Zip bytes cannot be empty")

        try:
            file_list = zip_bytes__file_list(request.zip_bytes)                        # Validate it's a valid zip
        except Exception as e:
            return Schema__Cache__Zip__Store__Response(success       = False                        ,
                                                       namespace     = request.namespace            ,
                                                       error_type    = "INVALID_ZIP_FORMAT"         ,
                                                       error_message = f"Invalid zip file: {str(e)}")

        hash_length = self.cache_service.hash_config.length
        cache_hash  = zip_bytes__content_hash(zip_bytes   = request.zip_bytes,
                                              hash_length = hash_length)                             # use the bytes content and files path to calculate the zip's hash (which is different mode than when we just store bytes)

        store_result = self.cache_service.store_with_strategy(storage_data     = request.zip_bytes    ,              # Store using cache service
                                                              cache_hash       = cache_hash           ,
                                                              cache_key        = request.cache_key    ,
                                                              file_id          = request.file_id      ,
                                                              namespace        = request.namespace    ,
                                                              strategy         = request.strategy     )

        if not store_result:                                                                                # Handle storage failure
            raise RuntimeError("Failed to store zip file in cache")

        return Schema__Cache__Zip__Store__Response(cache_id         = store_result.cache_id   ,             # Build response
                                                   cache_hash       = store_result.cache_hash ,
                                                   namespace        = store_result.namespace  ,
                                                   paths            = store_result.paths      ,
                                                   size             = store_result.size       ,
                                                   file_count       = len(file_list)          ,
                                                   success          = True              )


from typing                                                                       import Dict, Any, Optional
from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from osbot_utils.utils.Misc                                                       import timestamp_now
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                import CACHE__TEST__FIXTURES__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                         import Cache__Service

FILE_ID__CACHE__TEST__FIXTURES__MANIFEST = "00000000-0000-0000-0000-000000000001"

class Cache__Test__Fixtures(Type_Safe):                                                         # Manages reusable test fixtures for cache service tests
    cache_service     : Cache__Service                    = None                                # Cache service instance for fixture storage
    manifest_cache_id : Cache_Id                          = None                                # Predictable ID for manifest storage
    namespace         : Safe_Str__Id                      = CACHE__TEST__FIXTURES__NAMESPACE    # Namespace for fixture isolation
    fixtures          : Dict[str, dict[str, Any]]                                               # Map of fixture names to metadata
    setup_completed   : bool                              = False                               # Prevents redundant setup
    manifest          : Dict[str, Any]                    = None                                # Loaded manifest data
    delete_on_exit    : bool                              = False                               # Whether to cleanup fixtures

    # Core fixtures that tests commonly need
    default_fixtures = { 'string_simple'   : "test retrieve string data"                         ,
                         'string_medium'   : "test data " * 50                                   ,     # ~500 chars
                         'string_large'    : "x" * 1000                                          ,
                         'json_simple'     : {"key": "value", "number": 42}                      ,
                         'json_complex'    : { "nested" : {"data": [1, 2, 3], "more": {"level": 3}},
                                              "flag"    : True                                    ,
                                              "count"   : 100                                     ,
                                              "tags"    : ["test", "fixture", "cache"]           },
                         'json_empty'      : {}                                                  ,
                         'binary_small'    : b'\x01\x02\x03\x04\x05'                             ,
                         'binary_medium'   : bytes(range(100))                                   ,
                         'binary_large'    : bytes(range(256)) * 4                               }

    def setup(self) -> 'Cache__Test__Fixtures':                                       # Initialize cache service and load/create fixtures
        if self.setup_completed:
            return self

        if not self.namespace:                                                        # Set default namespace if not provided
            self.namespace = Safe_Str__Id("test-fixtures")

        if not self.manifest_cache_id:                                                # Use predictable GUID for manifest
            self.manifest_cache_id = Cache_Id(FILE_ID__CACHE__TEST__FIXTURES__MANIFEST)

        if self.load_manifest():                                                      # Try to load existing manifest
            #if self.verify_fixtures():
            self.setup_completed = True                                               # if manifest file exists, assume that all fixtures exist
            return self                                                               # i.e. there is no need to run this check on every request

        self.create_fixtures()                                                        # Create missing fixtures
        self.save_manifest()
        self.setup_completed = True
        return self

    def load_manifest(self) -> bool:                                                  # Load fixture manifest from cache service
        result = self.load_manifest__data()
        if result:
            self.manifest = result.get("data")
            self.fixtures = self.manifest.get("fixtures", {})
            return True
        return False

    def load_manifest__data(self):
        return self.cache_service.retrieve_by_id(cache_id  = self.manifest_cache_id,
                                                 namespace = self.namespace        )

    def verify_fixtures(self) -> bool:                                                # Check all fixtures still exist
        for fixture_name, fixture_info in self.fixtures.items():
            cache_id = Cache_Id(fixture_info.get("cache_id"))
            result = self.cache_service.retrieve_by_id(cache_id  = cache_id,
                                                       namespace = self.namespace)
            if not result:
                return False                                                           # Missing fixture, need to recreate
        return True

    def create_fixtures(self):                                                        # Create test fixtures in cache
        for name, data in self.default_fixtures.items():
            if name not in self.fixtures:
                if isinstance(data, bytes):                                           # Determine hash based on data type
                    cache_hash = self.cache_service.hash_from_bytes(data)
                elif isinstance(data, dict):
                    cache_hash = self.cache_service.hash_from_json(data)
                else:
                    cache_hash = self.cache_service.hash_from_string(str(data))

                cache_id = Cache_Id(Random_Guid())

                self.cache_service.store_with_strategy(storage_data = data        ,   # Store in cache
                                                       cache_hash   = cache_hash  ,
                                                       cache_id     = cache_id    ,
                                                       strategy     = "direct"     ,   # Use direct for test fixtures
                                                       namespace    = self.namespace)

                self.fixtures[name] = { "cache_id"   : str(cache_id)          ,         # Update fixtures map
                                        "cache_hash" : str(cache_hash)        ,
                                        "type"       : type(data).__name__    }

    def save_manifest(self):                                                          # Save fixture manifest to cache
        manifest_data = { "fixtures"   : self.fixtures     ,
                         "created_at" : timestamp_now()    }

        manifest_hash = self.cache_service.hash_from_json(manifest_data)

        self.cache_service.store_with_strategy(storage_data = manifest_data      ,
                                               cache_hash   = manifest_hash      ,
                                               cache_id     = self.manifest_cache_id,
                                               strategy     = "direct"            ,
                                               namespace    = self.namespace     )
        self.manifest = manifest_data

    def get_fixture(self, name: str) -> Optional[Dict[str, Any]]:                     # Get fixture info by name
        return self.fixtures.get(name)

    def get_fixture_data(self, name: str) -> Any:                                     # Retrieve actual fixture data
        fixture_info = self.get_fixture(name)
        if fixture_info:
            cache_id = Cache_Id(fixture_info["cache_id"])
            result = self.cache_service.retrieve_by_id(cache_id  = cache_id,
                                                       namespace = self.namespace)
            return result.get("data") if result else None
        return None

    def get_fixture_id(self, name: str) -> Optional[Cache_Id]:                     # Get cache ID for fixture
        fixture_info = self.get_fixture(name)
        if fixture_info:
            return Cache_Id(fixture_info["cache_id"])
        return None

    def get_fixture_hash(self, name: str) -> Optional[str]:                           # Get hash for fixture
        fixture_info = self.get_fixture(name)
        if fixture_info:
            return fixture_info["cache_hash"]
        return None

    def cleanup_fixtures(self):                                                       # Delete all fixtures (optional cleanup)
        if not self.delete_on_exit:
            return

        for fixture_name, fixture_info in self.fixtures.items():
            cache_id = Cache_Id(fixture_info["cache_id"])
            self.cache_service.delete_by_id(cache_id  = cache_id,
                                           namespace = self.namespace)

        self.cache_service.delete_by_id(cache_id  = self.manifest_cache_id,           # Delete manifest
                                        namespace = self.namespace        )

        self.fixtures.clear()
        self.manifest = {}

    def cleanup_all(self):      # Complete cleanup including bucket
        self.cleanup_fixtures()  # Delete fixtures



from fastapi                                                                                                 import Response
from typing                                                                                                      import List, Dict, Any
from osbot_fast_api.api.decorators.route_path                                                                import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                              import Fast_API__Routes
from osbot_utils.decorators.methods.cache_on_self                                                            import cache_on_self
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path                            import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now                             import Timestamp_Now
from osbot_utils.utils.Json                                                                              import bytes_to_json, json_to_str
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service                                      import DEFAULT__HTTP_CODE__FILE_NOT_FOUND
from mgraph_ai_service_cache.service.cache.Cache__Service                                                    import Cache__Service
from mgraph_ai_service_cache_client.schemas.routes.admin.Schema__Routes__Admin__Storage__Files_All__Response import Schema__Routes__Admin__Storage__Files_All__Response

TAG__ROUTES_STORAGE   = 'admin/storage'

# todo:review the use of namespace in path, since it creates an inconsistency without adding a lot of value
ROUTES_PATHS__STORAGE = [ f'/{TAG__ROUTES_STORAGE}/bucket-name'                          ,  # Current bucket name
                          f'/{TAG__ROUTES_STORAGE}/file/exists/{{path:path}}'            ,  # File Exists
                          f'/{TAG__ROUTES_STORAGE}/file/bytes/{{path:path}}'             ,  # File contents (as bytes)
                          f'/{TAG__ROUTES_STORAGE}/file/json/{{path:path}}'              ,  # File contents (as json)
                          f'/{TAG__ROUTES_STORAGE}/file/html/{{path:path}}'              ,  # File contents (as html)
                          f'/{TAG__ROUTES_STORAGE}/files/in/{{path:path}}'               ,  # Files in Path
                          f'/{TAG__ROUTES_STORAGE}/files/all/{{path:path}}'              ,  # Files all Path
                          f'/{TAG__ROUTES_STORAGE}/folders/{{path:path}}'                ,  # Folders
                          f'/{TAG__ROUTES_STORAGE}/file/delete/{{path:path}}'            ]  # File Delete

# todo: move to different Fast_API server/endpoint and add admin authorization
# todo: add missing type_safety return types
class Routes__Admin__Storage(Fast_API__Routes):
    tag           : str          = TAG__ROUTES_STORAGE
    cache_service : Cache__Service

    @cache_on_self
    def storage_fs(self):
        return self.cache_service.storage_fs()

    def bucket_name(self):
        storage = self.storage_fs()
        if hasattr(storage, 's3_bucket'):
            bucket_name = storage.s3_bucket
        else:
            bucket_name = 'NA'
        return {'bucket-name': bucket_name }

    @route_path("/file/exists/{path:path}")
    def file__exists(self, path):
        exists = self.storage_fs().file__exists(path)
        return dict(exists = exists,
                    path  =  path )

    @route_path("/file/bytes/{path:path}")
    def file__bytes(self, path):
        file_bytes = self.storage_fs().file__bytes(path)
        return Response(file_bytes, media_type='application/octet-stream')

    @route_path("/file/json/{path:path}")
    def file__json(self, path):
        file_bytes = self.storage_fs().file__bytes(path)
        if file_bytes:
            file_json   = bytes_to_json(file_bytes)
            return file_json
        error_data    = dict( error_type = "FILE_NOT_FOUND" ,
                              resource   = "file"           ,
                              message    = "The requested file does not exist in storage",
                              path       = path             )

        return Response(json_to_str(error_data)                           ,
                        status_code = DEFAULT__HTTP_CODE__FILE_NOT_FOUND  ,          # using 404 to indicate file not found
                        media_type  = 'application/json')

    @route_path("/file/html/{path:path}")
    def file__html(self, path):
        file_json = self.storage_fs().file__json(path)
        if file_json:
            html = file_json.get('html')
            return Response(html,
                            status_code = 200   ,
                            media_type  = "text/html; charset=utf-8")
        else:
            return Response('File not found.', status_code = 404   ,
                             media_type  = "text/html; charset=utf-8")


    @route_path("/files/in/{path:path}")
    def files__in(self,                                               # List files in path - optionally recursive
                  path            : Safe_Str__File__Path = ''   ,
                  return_full_path: bool                 = False ,
                  recursive       : bool                 = False
              ) -> List:

        if recursive:                                                                                       # Use recursive method - returns all files under path
            files = self.storage_fs().folder__files__all(parent_folder=path)                                # Note: folder__files__all always returns full paths
            return sorted(files)
        else:
            files = self.storage_fs().folder__files(folder_path=path, return_full_path=return_full_path)    # Use non-recursive method - returns files in directory only
            return files

    # todo: as it is this could have major performance implications (for large namespaces
    #       this is also a good example of the kind of method/capability that we could run on a schedule and return the cached version
    # todo: see if we need this since we can also get all files recursively using files__in__path
    @route_path("/files/all/{path:path}")
    def files__all(self,
                   path     : Safe_Str__File__Path = ''
              ) -> Schema__Routes__Admin__Storage__Files_All__Response:
        files      = self.storage_fs().folder__files__all(parent_folder=path)
        file_count = len(files)
        return Schema__Routes__Admin__Storage__Files_All__Response(timestamp  = Timestamp_Now(),
                                                                   file_count = file_count     ,
                                                                   files      = sorted(files)  )

    # todo: move the logic of this folders_path into a services method (since this type of logic should not be in the Routes class
    @route_path('/folders/{path:path}')
    def folders(self,
                path            : Safe_Str__File__Path = ''    ,                                    # List folders in path - optionally recursive
                return_full_path: bool                 = False ,
                recursive       : bool                 = False
          ) -> List:
        if recursive:                                                                                           # Recursive: get all folders under path
            all_folders = []
            folders = self.storage_fs().folder__folders(parent_folder=path, return_full_path=True)
            all_folders.extend(folders)

            for folder in folders:                                                                              # Recursively get subfolders
                subfolders = self.folders(path=folder, return_full_path=True, recursive=True)
                all_folders.extend(subfolders)

            result = sorted(list(set(all_folders)))                                                               # Remove duplicates and sort
            if not return_full_path and path:
                base_path = path if path.endswith('/') else path + '/'                                                # Remove the base path prefix from all results
                result = [folder.replace(base_path, '', 1) if folder.startswith(base_path) else folder
                          for folder in result ]
            return result
        else:
            folders = self.storage_fs().folder__folders(parent_folder=path, return_full_path=return_full_path)  # Non-recursive: get folders in this directory only
            return folders

    @route_path('/file/delete/{path:path}')
    def file__delete(self,
                     path : Safe_Str__File__Path
                ) -> Dict:
        result = self.cache_service.storage_fs().file__delete(path)
        return dict(path    = path,
                    deleted = result )


    def setup_routes(self):
        self.add_route_get   (self.bucket_name  )
        self.add_route_get   (self.file__exists )
        self.add_route_get   (self.file__bytes  )
        self.add_route_get   (self.file__json   )
        self.add_route_get   (self.file__html   )
        self.add_route_get   (self.files__in    )
        self.add_route_get   (self.files__all   )
        self.add_route_get   (self.folders      )
        self.add_route_delete(self.file__delete )



from fastapi                                                                                    import HTTPException, Response, Body
from osbot_fast_api.api.decorators.route_path                                                   import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                                 import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix                      import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag                         import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                               import cache_on_self
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                              import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id                 import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path               import Safe_Str__File__Path
from osbot_utils.utils.Zip                                                                      import zip_bytes_empty
from mgraph_ai_service_cache_client.schemas.cache.consts__Cache_Service                         import DEFAULT_CACHE__ZIP__STRATEGY
from mgraph_ai_service_cache_client.schemas.cache.enums.Enum__Cache__Store__Strategy            import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                              import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                                       import Cache__Service
from mgraph_ai_service_cache.service.cache.zip.Cache__Service__Zip__Store                       import Cache__Service__Zip__Store
from mgraph_ai_service_cache.service.cache.zip.Cache__Service__Zip__Operations                  import Cache__Service__Zip__Operations
from mgraph_ai_service_cache.service.cache.zip.Cache__Service__Zip__Batch                       import Cache__Service__Zip__Batch
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Store__Request        import Schema__Cache__Zip__Store__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Store__Response       import Schema__Cache__Zip__Store__Response
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Operation__Request    import Schema__Cache__Zip__Operation__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Operation__Response   import Schema__Cache__Zip__Operation__Response
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Batch__Request        import Schema__Cache__Zip__Batch__Request
from mgraph_ai_service_cache_client.schemas.cache.zip.Schema__Cache__Zip__Batch__Response       import Schema__Cache__Zip__Batch__Response


TAG__ROUTES_ZIP    = Safe_Str__Fast_API__Route__Tag('zip')
PREFIX__ROUTES_ZIP = Safe_Str__Fast_API__Route__Prefix('/{namespace}')
BASE_PATH__ZIP     = f'{PREFIX__ROUTES_ZIP}/{TAG__ROUTES_ZIP}'

ROUTES_PATHS__ZIP = [ f'{PREFIX__ROUTES_ZIP}/{{strategy}}/zip/create/{{cache_key:path}}/{{file_id}}'        ,    # create new zip (empty)
                      f'{PREFIX__ROUTES_ZIP}/{{strategy}}/zip/store/{{cache_key:path}}/{{file_id}}'          ,    # Store new zip
                      f'{BASE_PATH__ZIP}/{{cache_id}}/files/list'                                       ,    # List files in zip
                      f'{BASE_PATH__ZIP}/{{cache_id}}/file/retrieve/{{file_path:path}}'                 ,    # Get specific file from zip
                      f'{BASE_PATH__ZIP}/{{cache_id}}/file/add/from/string/{{file_path:path}}'          ,    # Add file from string
                      f'{BASE_PATH__ZIP}/{{cache_id}}/file/add/from/bytes/{{file_path:path}}'           ,    # Add file from bytes
                      f'{BASE_PATH__ZIP}/{{cache_id}}/file/delete/{{file_path:path}}'                   ,    # Delete file from zip
                      f'{BASE_PATH__ZIP}/{{cache_id}}/batch/operations'                                 ,    # Batch operations
                      f'{BASE_PATH__ZIP}/{{cache_id}}/retrieve'                                         ]    # Retrieve entire zip

class Routes__Zip(Fast_API__Routes):                                       # FastAPI routes for zip operations
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_ZIP
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_ZIP
    cache_service : Cache__Service                                         # Dependency injection

    @cache_on_self
    def zip_store_service(self) -> Cache__Service__Zip__Store:             # Service for storing zips
        return Cache__Service__Zip__Store(cache_service=self.cache_service)

    @cache_on_self
    def zip_ops_service(self) -> Cache__Service__Zip__Operations:          # Service for zip operations
        return Cache__Service__Zip__Operations(cache_service=self.cache_service)

    @cache_on_self
    def zip_batch_service(self) -> Cache__Service__Zip__Batch:             # Service for batch operations
        return Cache__Service__Zip__Batch(cache_service=self.cache_service)

    @route_path("/{strategy}/zip/create/{cache_key:path}/{file_id}")
    def zip_create(self,namespace  : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE  ,
                        strategy   : Enum__Cache__Store__Strategy = DEFAULT_CACHE__ZIP__STRATEGY,
                        cache_key  : Safe_Str__File__Path         = None,
                        file_id    : Safe_Str__Id                 = None
                   ) -> Schema__Cache__Zip__Store__Response:                # Store a new zip file

        zip_bytes = zip_bytes_empty()
        request = Schema__Cache__Zip__Store__Request(zip_bytes = zip_bytes        ,
                                                     cache_key = cache_key   ,
                                                     file_id   = file_id     ,
                                                     namespace = namespace   ,
                                                     strategy = strategy     )
        return self.zip_store_service().store_zip(request)


    @route_path("/{strategy}/zip/store/{cache_key:path}/{file_id}")
    def zip_store(self, body       : bytes = Body(..., media_type="application/zip"),
                        namespace  : Safe_Str__Id                 = FAST_API__PARAM__NAMESPACE     ,
                        strategy   : Enum__Cache__Store__Strategy = DEFAULT_CACHE__ZIP__STRATEGY   ,
                        cache_key  : Safe_Str__File__Path         = None,
                        file_id    : Safe_Str__Id                 = None
                   ) -> Schema__Cache__Zip__Store__Response:                # Store a new zip file

        request = Schema__Cache__Zip__Store__Request(zip_bytes = body        ,
                                                     cache_key = cache_key   ,
                                                     file_id   = file_id     ,
                                                     strategy  = strategy    ,
                                                     namespace = namespace   )
        response = self.zip_store_service().store_zip(request)
        if not response.success:                                                        # Route decides the HTTP semantics
            if response.error_type == "INVALID_INPUT":                                  # todo: move these strings to Enum
                raise HTTPException(status_code=400, detail=response.error_message)
            elif response.error_type == "INVALID_ZIP_FORMAT":
                raise HTTPException(status_code=400, detail=response.error_message)
            else:
                raise HTTPException(status_code=500, detail=response.error_message)
        return response

    @route_path("/zip/{cache_id}/files/list")
    def zip_files_list(self, cache_id : Cache_Id   ,
                             namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                        ) -> Schema__Cache__Zip__Operation__Response:       # List files in stored zip

        request = Schema__Cache__Zip__Operation__Request(cache_id  = cache_id   ,
                                                         operation = "list"     ,
                                                         namespace = namespace  )

        result = self.zip_ops_service().perform_operation(request)

        if not result.success:
            raise HTTPException(status_code=404, detail=result.error_details)

        return result

    @route_path("/zip/{cache_id}/file/retrieve/{file_path:path}")
    def zip_file_retrieve(self, cache_id  : Cache_Id   ,
                           file_path : Safe_Str__File__Path,
                           namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                      ) -> Response:                                         # Get specific file from zip

        request = Schema__Cache__Zip__Operation__Request(cache_id  = cache_id   ,
                                                         operation = "get"      ,
                                                         file_path = file_path  ,
                                                         namespace = namespace  )

        result = self.zip_ops_service().perform_operation(request)

        if not result.success:
            raise HTTPException(status_code=404, detail=result.error_details)

        return Response(content=result.file_content, media_type="application/octet-stream")

    @route_path("/zip/{cache_id}/file/add/from/string/{file_path:path}")
    def zip_file_add_from_string(self, cache_id  : Cache_Id   ,
                                       body      : str = Body(...),
                                       file_path : Safe_Str__File__Path = None,
                                       namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                  ) -> Schema__Cache__Zip__Operation__Response:                 # Add file to zip (from string)
        body_as_bytes = body.encode("utf-8")
        return self.zip_file_add_from_bytes(cache_id = cache_id  ,
                                            body = body_as_bytes ,
                                            file_path = file_path ,
                                            namespace = namespace ,)

    @route_path("/zip/{cache_id}/file/add/from/bytes/{file_path:path}")
    def zip_file_add_from_bytes(self, cache_id  : Cache_Id   ,
                                      body      : bytes = Body(...),
                                      file_path : Safe_Str__File__Path = None,
                                      namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                                 ) -> Schema__Cache__Zip__Operation__Response:         # Add file to zip

        if not file_path:
            raise HTTPException(status_code=400, detail="file_path required")

        request = Schema__Cache__Zip__Operation__Request(cache_id     = cache_id   ,
                                                         operation    = "add"      ,
                                                         file_path    = file_path  ,
                                                         file_content = body       ,
                                                         namespace    = namespace)

        result = self.zip_ops_service().perform_operation(request)

        if not result.success:
            raise HTTPException(status_code=400, detail=result.error_details)

        return result

    @route_path("/zip/{cache_id}/file/delete/{file_path:path}")
    def zip_file_delete(self, cache_id  : Cache_Id   ,
                              file_path : Safe_Str__File__Path,
                              namespace : Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                         ) -> Schema__Cache__Zip__Operation__Response:      # Remove file from zip

        request = Schema__Cache__Zip__Operation__Request(cache_id  = cache_id   ,
                                                         operation = "remove"   ,
                                                         file_path = file_path  ,
                                                         namespace = namespace  )

        result = self.zip_ops_service().perform_operation(request)

        if not result.success:
            raise HTTPException(status_code=400, detail=result.error_details)

        return result

    @route_path("/zip/{cache_id}/batch/operations")
    def batch_operations(self, request  : Schema__Cache__Zip__Batch__Request   ,
                               cache_id : Cache_Id                          ,
                               namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                          ) -> Schema__Cache__Zip__Batch__Response:         # Perform batch operations

        request.cache_id  = cache_id                                        # Ensure cache_id and namespace match  #todo: review this pattern, since it very weird to accept a value that is not used (in this case request.cache_id)
        request.namespace = namespace
        try:
            return self.zip_batch_service().perform_batch(request)
        except Exception as e:
            raise HTTPException(status_code=500, detail=f"Batch operation failed: {str(e)}")

    @route_path("/zip/{cache_id}/retrieve")
    def zip_retrieve(self, cache_id : Cache_Id   ,
                           namespace: Safe_Str__Id = FAST_API__PARAM__NAMESPACE
                      ) -> Response:                                         # Download entire zip file

        result = self.cache_service.retrieve_by_id(cache_id, namespace)

        if not result:
            raise HTTPException(status_code=404, detail="Zip file not found")

        if result.get('data_type') != 'binary':
            raise HTTPException(status_code=400, detail="Cached item is not a binary file")

        zip_bytes = result.get('data')

        return Response(
            content=zip_bytes,
            media_type="application/zip",
            headers={"Content-Disposition": f"attachment; filename={cache_id}.zip"}
        )

    def setup_routes(self):                                                # Configure all routes
        self.add_route_post  (self.zip_create              )
        self.add_route_post  (self.zip_store               )
        self.add_route_get   (self.zip_retrieve            )
        self.add_route_get   (self.zip_files_list          )
        self.add_route_get   (self.zip_file_retrieve       )
        self.add_route_post  (self.zip_file_add_from_bytes )
        self.add_route_post  (self.zip_file_add_from_string)
        self.add_route_delete(self.zip_file_delete         )
        self.add_route_post  (self.batch_operations        )




from typing                                                                          import Any
from fastapi                                                                         import HTTPException, Body
from osbot_fast_api.api.decorators.route_path                                        import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                      import Fast_API__Routes
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Prefix           import Safe_Str__Fast_API__Route__Prefix
from osbot_fast_api.api.schemas.safe_str.Safe_Str__Fast_API__Route__Tag              import Safe_Str__Fast_API__Route__Tag
from osbot_utils.decorators.methods.cache_on_self                                    import cache_on_self
from osbot_utils.type_safe.primitives.domains.identifiers.Cache_Id                   import Cache_Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id      import Safe_Str__Id
from osbot_utils.type_safe.type_safe_core.decorators.type_safe                       import type_safe
from mgraph_ai_service_cache_client.schemas.cache.Schema__Cache__Update__Response    import Schema__Cache__Update__Response
from mgraph_ai_service_cache_client.schemas.consts.const__Fast_API                   import FAST_API__PARAM__NAMESPACE
from mgraph_ai_service_cache.service.cache.Cache__Service                            import Cache__Service
from mgraph_ai_service_cache.service.cache.update.Cache__Service__Update             import Cache__Service__Update

TAG__ROUTES_UPDATE                  = 'update'
PREFIX__ROUTES_UPDATE               = '/{namespace}'
BASE_PATH__ROUTES_UPDATE            = f'{PREFIX__ROUTES_UPDATE}/update/{{cache_id}}'
ROUTES_PATHS__UPDATE                = [ BASE_PATH__ROUTES_UPDATE + '/string'   ,
                                        BASE_PATH__ROUTES_UPDATE + '/json'     ,
                                        BASE_PATH__ROUTES_UPDATE + '/binary'   ]


class Routes__File__Update(Fast_API__Routes):                                   # FastAPI routes for updating existing cache entries
    tag           : Safe_Str__Fast_API__Route__Tag    = TAG__ROUTES_UPDATE
    prefix        : Safe_Str__Fast_API__Route__Prefix = PREFIX__ROUTES_UPDATE
    cache_service : Cache__Service                                              # Dependency injection for cache service

    @cache_on_self
    def update_service(self) -> Cache__Service__Update:                         # Service layer for update operations
        return Cache__Service__Update(cache_service=self.cache_service)

    @route_path("/update/{cache_id}/string")
    def update__string(self,
                       data      : str           = Body(...)                     ,
                       cache_id  : Cache_Id      = None                          ,
                       namespace : Safe_Str__Id  = FAST_API__PARAM__NAMESPACE
                  ) -> Schema__Cache__Update__Response:                          # Update existing cache entry with string data

        if not data:                                                            # Validate input
            error_detail = { "error_type"    : "INVALID_INPUT"                       ,
                             "message"       : "String data cannot be empty"         ,
                             "field_name"    : "data"                                }
            raise HTTPException(status_code=400, detail=error_detail)

        return self._execute_update(cache_id  = cache_id  ,
                                    namespace = namespace ,
                                    data      = data      )

    @route_path("/update/{cache_id}/json")
    def update__json(self,
                     data      : dict          = Body(...)                     ,
                     cache_id  : Cache_Id      = None                          ,
                     namespace : Safe_Str__Id  = FAST_API__PARAM__NAMESPACE
                ) -> Schema__Cache__Update__Response:                            # Update existing cache entry with JSON data

        return self._execute_update(cache_id  = cache_id  ,
                                    namespace = namespace ,
                                    data      = data      )

    @route_path("/update/{cache_id}/binary")
    def update__binary(self,
                       body      : bytes         = Body(..., media_type="application/octet-stream"),
                       cache_id  : Cache_Id      = None                          ,
                       namespace : Safe_Str__Id  = FAST_API__PARAM__NAMESPACE
                  ) -> Schema__Cache__Update__Response:                          # Update existing cache entry with binary data

        if not body:                                                            # Validate input
            error_detail = { "error_type"    : "INVALID_INPUT"                       ,
                             "message"       : "Binary data cannot be empty"         ,
                             "field_name"    : "body"                                }
            raise HTTPException(status_code=400, detail=error_detail)

        return self._execute_update(cache_id  = cache_id  ,
                                    namespace = namespace ,
                                    data      = body      )

    @type_safe
    def _execute_update(self,
                        cache_id  : Cache_Id      ,
                        namespace : Safe_Str__Id  ,
                        data      : Any           ,
                   ) -> Schema__Cache__Update__Response:                         # Common update logic for all data types

        result = self.update_service().update_by_id(cache_id  = cache_id  ,
                                                    namespace = namespace ,
                                                    data      = data      )

        if result is None:                                                      # Handle update failure
            error_detail = { "error_type" : "UPDATE_FAILED"                          ,
                             "message"    : f"Failed to update cache entry {cache_id}",
                             "cache_id"   : str(cache_id)                            }
            raise HTTPException(status_code=500, detail=error_detail)

        return result

    def setup_routes(self):                                                     # Configure all update routes
        self.add_route_post(self.update__string)
        self.add_route_post(self.update__json  )
        self.add_route_post(self.update__binary)